/*
  OMPi OpenMP Compiler
  == Copyright since 2001 the OMPi Team
  == Department of Computer Science, University of Ioannina

  This file is part of OMPi.

  OMPi is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  OMPi is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with OMPi; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* x_for.c */

/*
 * 2009/05/11:
 *   added AUTO schedule type - implemented as STATIC for now.
 */


#include <string.h>
#include "x_for.h"
#include "x_clauses.h"
#include "ast_xform.h"
#include "ast_free.h"
#include "ast_copy.h"
#include "ast_print.h"
#include "str.h"
#include "ompi.h"

/* Possible clauses:
 * private, firstprivate, lastprivate, reduction, nowait, ordered, schedule
 */
void xform_for(aststmt *t)
{
  aststmt   s = (*t)->u.omp->body, parent = (*t)->parent, v,
            decls, inits = NULL, lasts = NULL, reds = NULL, stmp;
  astexpr   cond = s->u.iteration.cond,
            incr = s->u.iteration.incr,
            lb, b, step;
  symbol    var;
  int       condop, incrop;
  int       haslast, hasboth, hasred;
  int       schedtype = OC_static /* default */, 
            static_chunk = 0;
  astexpr   schedchunk = NULL;
  ompclause nw = xc_ompcon_get_clause((*t)->u.omp, OCNOWAIT),
            sch = xc_ompcon_get_clause((*t)->u.omp, OCSCHEDULE),
            ord = xc_ompcon_get_clause((*t)->u.omp, OCORDERED);
  symtab    dvars;
  str       slb = Strnew(), sb = Strnew(), sstep = Strnew(), schsize = Strnew();
  stentry   varentry;
  
  v = ompdir_commented((*t)->u.omp->directive); /* Put directive in comments */

  /*
   * Preparations
   */

     /* Gather all info for the FOR statement of the body */
  if (s->u.iteration.init->type == EXPRESSION)
  {
    astexpr init = s->u.iteration.init->u.expr;   /* Assignment var = lb */
    var = init->left->u.sym;
    lb  = init->right;
  }
  else    /* Declaration */
  {
    lb  = s->u.iteration.init->u.declaration.decl->u.expr; /* type var = lb */
    var = s->u.iteration.init->u.declaration.decl->decl->decl->u.id;
  }
  condop = cond->opid;       
  b      = cond->right;
  incrop = incr->right->opid;
  step   = incr->right->right;   /* if incrop is -, step must change sign! */
     /* Sanity check */
  if (var != cond->left->u.sym || var != incr->left->u.sym)
    exit_error(1, "(%s, line %d) openmp error:\n\t"
      "a single iteration variable should be used in the FOR statement.\n",
      (*t)->u.omp->directive->file->name,(*t)->u.omp->directive->l);
     /* Fix the upper bound; we don't like <= / >= */
  if (condop == BOP_leq || condop == BOP_geq)
  {
    /* [from original OMPi:]
     * Take care of <= or >=, by adding or subtracting 1.
     * Our runtime code assumes pure < / >.
     * Clearly, this breaks down if the loop's upper bound is MAXINT :-(.
     */
    condop = (condop == BOP_leq) ? BOP_lt : BOP_gt;
    b = BinaryOperator((condop == BOP_lt) ? BOP_add : BOP_sub, 
                       UnaryOperator(UOP_paren, b), 
                       Constant(strdup("1")));
  }
  ast_expr_print(sb,    b);          /* Store in strings, too */
  ast_expr_print(slb,   lb);
  ast_expr_print(sstep, step);
     /* Know thy schedule */
  if (sch)
  {
    if (sch->subtype == OC_auto)      /* Implement AUTO as STATIC for now */
      schedtype = OC_static;          /* but we will ignore any chunksize */
    else
      schedtype  = sch->subtype;      /* OC_static, OC_... */
    schedchunk = sch->u.expr;
    if (schedtype == OC_static && sch->subtype != OC_auto && schedchunk)
      static_chunk = 1;
    if (schedtype == OC_affinity && schedchunk)
      schedchunk = ast_expr_copy(schedchunk);
    if (schedchunk)
      ast_expr_print(schsize, schedchunk);
  }
     /* Collect all data clause vars - we need to check if any vars
      * are both firstprivate and lastprivate 
      */  
  dvars = xc_validate_store_dataclause_vars((*t)->u.omp->directive);
     /* Must check if the loop variable has been enlisted; if not,
      * it is automatically considered private (v25) - so we make it
      * appear as if there was a private(var) clause.
      */
  if ((varentry = symtab_get(dvars, var, IDNAME)) == NULL)
  {
    if (s->u.iteration.init->type == EXPRESSION)  /* if DECLARATION, leave it */
      symtab_put(dvars, var, IDNAME)->ival = OCPRIVATE;
  }
  else
  {
    if (s->u.iteration.init->type != EXPRESSION)  /* a declaration */
      exit_error(1, "(%s, line %d) openmp error:\n\t"
        "iteration variable '%s' is declared within the FOR statement\n\t"
        "and thus it cannot appear in the directive's data clauses.\n",
        (*t)->u.omp->directive->file->name,(*t)->u.omp->directive->l,var->name);
    /* Remove the FIRSTPRIVATE attribute if any (there is no use for it) */
    /* Actually, v25 (p.64,l.23) specifies that the iteration variable
     * can only appear in a PRIVATE or LASTPRIVATE clause, so we should
     * emit at least a warning.
     */
    if (varentry->ival==OCFIRSTPRIVATE || varentry->ival==OCFIRSTLASTPRIVATE)
      warning("(%s, line %d) warning:\n\t"
        "iteration variable '%s' cannot appear in a FIRSTPRIVATE clause..\n\t"
        "  .. let's pretend it was in a PRIVATE clause.\n",
        (*t)->u.omp->directive->file->name,(*t)->u.omp->directive->l,var->name);
    if (varentry->ival == OCFIRSTPRIVATE)
      varentry->ival = OCPRIVATE;
    else
      if (varentry->ival == OCFIRSTLASTPRIVATE)
        varentry->ival = OCLASTPRIVATE;
  }
  
     /* declarations from the collected vars (not the clauses!) */
  decls = xc_stored_vars_declarations(&haslast, &hasboth, &hasred);
     /* initialization statments for firstprivate non-scalar vars */
  if (decls)
    inits = xc_ompdir_fiparray_initializers((*t)->u.omp->directive);
     /* assignments for lastprivate vars */
  if (haslast)
    lasts = xc_ompdir_lastprivate_assignments((*t)->u.omp->directive);
  if (hasred)
    reds = xc_ompdir_reduction_code((*t)->u.omp->directive);
     /* we need 4 or 5 more variables */
  stmp = (schedtype == OC_affinity) ?
           ( (haslast) ?
               parse_blocklist_string("int __lidx, last_iter_=0;"
                                      "struct _ort_gdopt_ gdopt_;") :
               parse_blocklist_string("struct _ort_gdopt_ gdopt_;")
          ) :
           ( (haslast) ?
               parse_blocklist_string("int from_=0, to_=0, step_, last_iter_=0;"
                                      "struct _ort_gdopt_ gdopt_;") :
               parse_blocklist_string("int from_=0, to_=0, step_;"
                                      "struct _ort_gdopt_ gdopt_;")
           );
     /* we may need 4 more variables */
  if (static_chunk)
    stmp = BlockList(
             stmp,
             parse_blocklist_string("int nchunks_, chid_, TN_, StCtN_;")
           );
     /* we may need 2 more variables */
  if (schedtype == OC_runtime)
    stmp = BlockList(
             stmp,
             parse_blocklist_string("int staticextra_ = 0, chunksize_; "
                                    "chunky_t get_chunk_;")
           );
  decls = (decls) ? BlockList(decls,stmp) : stmp;     
     /* checks for ditching the implicit barrier */
  if (nw == NULL && !xform_implicit_barrier_is_needed((*t)->u.omp))
    nw = (ompclause) 123;       /* Non-null anyway */
  
  /*
   * Do the job
   */

  (*t)->u.omp->body = NULL;     /* Make it NULL so as to free it easily */
  ast_free(*t);                 /* Get rid of the OmpStmt */
  
  if (inits)
    decls = BlockList(decls, inits);  /* Append the initialization statements */
  
  /* Append our new code: step_ = <step>; ort_entering_for(...); */
  if (schedtype == OC_affinity)
  {
    stmp = (haslast) ?
             BlockList(
               decls,
               parse_blocklist_string("__lidx = %s + (((%s)-(%s))/(%s))*(%s); "
                                     "if (__lidx %c= %s) __lidx %c= %s; ",
                                     str_string(slb), str_string(sb), 
                                     str_string(slb), str_string(sstep), 
                                     str_string(sstep), 
                                     (condop == BOP_lt) ? '>' : '<', 
                                     str_string(sb), 
                                     (condop == BOP_lt) ? '-' : '+', 
                                     str_string(sstep))
             ) :
             decls;
    stmp = BlockList(
             stmp,
             parse_blocklist_string("ort_entering_for(%d,%d,%s,%s,&gdopt_);",
                                    nw ? 1 : 0, ord ? 1 : 0, 
                                    str_string(slb),str_string(sstep))
           );
  }
  else
    stmp = BlockList(
             decls,
             parse_blocklist_string("step_ =%c%s;"
                                    "ort_entering_for(%d,%d,%s,step_,&gdopt_);",
                                    incrop == BOP_add ? ' ' : '-',
                                    str_string(sstep), nw ? 1 : 0, ord ? 1 : 0, 
                                    str_string(slb))
           );
  if (hasboth)   /* a var is both fip & lap; this needs a barrier here :-( */
    stmp = BlockList(stmp, BarrierCall());

    /* Prepare the main loop */
  if (schedtype == OC_affinity)
  {
      /* same loop, new body */
    if (lasts || ord)
    {
      if (lasts)
        s->body = BlockList(
                    parse_blocklist_string(
                      "if (%s == __lidx) last_iter_ = 1;", var->name
                    ),
                    s->body
                  );
      if (ord)
        s->body = BlockList(
                    parse_blocklist_string("ort_thischunk_lb(%s);", var->name),
                    s->body
                  );
       s->body = Compound(s->body);
    }
    s->body = If(
                FunctionCall(
                  Identifier(Symbol("ort_affine_iteration")),schedchunk
                ),
                s->body,
                NULL
              );
  }
  else
  {        
      /* new bounds, same body: for (var = from_; var </> to_; <incr>) <body> */
    s = For((s->u.iteration.init->type == EXPRESSION) ?
              Expression(Assignment(Identifier(var), /* <var> = from_; */
                                    ASS_eq,
                                    Identifier(Symbol("from_")))
              )
               :
              Declaration(
                 ast_spec_copy_nosc(s->u.iteration.init->u.declaration.spec),
                 InitDecl(
                   Declarator(NULL, IdentifierDecl(var)),
                   Identifier(Symbol("from_"))
                 )
               ),
            BinaryOperator(condop, Identifier(var), Identifier(Symbol("to_"))),
            incr,
            /* TODO: put # line info here */
            s->body);
        /* The ordered_begin() needs to know the current chunk's start
         * (from_) in order to work. However, "from_" may be unavailable
         * as a variable if the ORDERED directive is orphaned. Thus,
         * we have to store is somewhere just in case.
         */
    if (ord)
      s = BlockList(parse_blocklist_string("ort_thischunk_lb(from_);"), s);
    if (lasts)              /* if (from_ </> sb && to_ == sb) last_iter = 1; */
      s = BlockList(
            parse_blocklist_string(
              "if (from_ %c (%s) && to_ == (%s)) last_iter_ = 1;",
              (condop == BOP_lt) ? '<' : '>', str_string(sb), str_string(sb)),
            s
          );
  }

  /* Schedule-dependent code */
  switch (schedtype)
  {
    case OC_static:
      if (static_chunk)
        stmp = BlockList(
                 BlockList(
                   stmp,
                   parse_blocklist_string(
                     "TN_ = omp_get_num_threads();"
                     "StCtN_ = step_*%s*TN_;"
                   "ort_init_static_chunksize(%s,%s,step_,%s,&nchunks_,&from_);"
                     "from_ -= StCtN_;"
                     "to_ = from_ + step_*(%s);",
                     str_string(schsize),  
                     str_string(slb), str_string(sb), str_string(schsize),
                     str_string(schsize)
                   )
                 ),
                 For(
                   parse_blocklist_string("chid_ = omp_get_thread_num();"),
                   parse_expression_string("chid_ < nchunks_"),
                   parse_expression_string("chid_ += TN_"),
                   Compound(
                     BlockList(
                       parse_blocklist_string(
                         "from_ += StCtN_;"
                         "to_ += StCtN_;"
                         "if (%s %c to_) to_ = %s;",
                         str_string(sb), 
                         (condop == BOP_lt) ? '<' : '>',
                         str_string(sb)
                       ),
                       s
                     )
                   )
                 )
               );
      else
        stmp = BlockList(
                 stmp,
                 If(
                   parse_expression_string(
                    "ort_get_static_default_chunk(%s, %s, step_, &from_, &to_)",
                    str_string(slb), str_string(sb)),
                   Compound(s), 
                   NULL
                 )
               );
      break;             

    case OC_guided:
    case OC_dynamic:
      stmp = BlockList(
               stmp,
               While(
                 parse_expression_string(
                 "ort_get_%s_chunk(%s,%s,step_,%s,&from_,&to_,(int*)0,&gdopt_)",
                   schedtype == OC_guided ? "guided" : "dynamic", 
                   str_string(slb), str_string(sb), 
                   schedchunk ? str_string(schsize) : "1"),
                 Compound(s)
               )
             );
      break;

    case OC_runtime:
      stmp = BlockList(
               BlockList(
                 stmp,
                 /* ort_get_runtime_schedule_stuff(&get_chunk_, &chunksize_); */
                 Expression(
                   FunctionCall( 
                     Identifier(Symbol("ort_get_runtime_schedule_stuff")), 
                     CommaList(
                       UnaryOperator(UOP_addr,Identifier(Symbol("get_chunk_"))),
                       UnaryOperator(UOP_addr,Identifier(Symbol("chunksize_")))
                     )
                   )
                 )
               ),
               While(
                 parse_expression_string(  /* Too big to do it by hand */
                   "(*get_chunk_)(%s, %s, step_, chunksize_,"
                   "&from_, &to_, &staticextra_, &gdopt_)",
                   str_string(slb), str_string(sb)
                 ),
                 Compound(s)
               )
             );
      break;

    case OC_affinity:
      stmp = BlockList(stmp, s);      
      break;
  }

  *t = BlockList(
         BlockList(v, stmp),         /* ort_leaving_for(); */
         Call0_stmt("ort_leaving_for")
       );
  if (lasts)
    *t = BlockList(
           *t,
           If(
             Identifier(Symbol("last_iter_")),
             lasts->type == STATEMENTLIST ?  Compound(lasts) : lasts,
             NULL
           )
         );
  if (reds)
    *t = BlockList(*t, reds);
  if (nw == NULL)                  /* Must output a barrier */
    *t = BlockList(*t, BarrierCall());
  
  *t = Compound(*t);
  ast_stmt_parent(parent, *t);
  
  str_free(slb);
  str_free(sb);
  str_free(sstep);
  str_free(schsize);
}
