/*
  OMPi OpenMP Compiler
  == Copyright since 2001 the OMPi Team
  == Department of Computer Science, University of Ioannina

  This file is part of OMPi.

  OMPi is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  OMPi is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with OMPi; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
 * An implementation of the othr API using PM2 marcel threads.
 * A general purpose two level thread library for nested parallelism.
 */



#include "../ort.h"
#include "ee.h"
#include <sched.h>
#include <stdlib.h>
#include <string.h>

/* Global stuff for the team */
typedef struct thr_info {
  marcel_t thr;
  void *team_argument;            /* The function's argument */
  int thr_id;
} thr_info_t;

struct team_info {
  int  team_nthr;                 /* # threads in the current team */
  thr_info_t *team_thr;
};

static int     marcel_lib_inited = 0;     /* Flag to avoid re-initializations */
marcel_attr_t mattr;
marcel_mutex_t at_lock;
unsigned long activethreads = 1;
unsigned long threadlimit;

#if !defined(MARCEL_KEYS_ENABLED)
#error "Marcel keys not enabled!"
#endif


/* Library initialization*/
int othr_initialize(int *argc, char ***argv, ort_icvs_t *icv, ort_caps_t *caps)
{
  int         nthr, i, e;
  void        *threadjob(void *);



  marcel_init(argc, *argv);
  nthr = (icv->nthreads >= 0) ?  /* Explicitely requested population */
               icv->nthreads :
               icv->ncpus - 1;   /* we don't handle the initial thread */
  caps->supports_nested            = 1;
  caps->supports_dynamic           = 1;
  caps->supports_nested_nondynamic = 1;
  caps->max_levels_supported       = -1;     /* No limit */
  caps->default_numthreads         = nthr;
  caps->max_threads_supported      = -1;     /* No limit */
  
  if ( marcel_lib_inited ) return (0);

  marcel_lib_inited = 1;
  marcel_attr_init(&mattr);
  marcel_mutex_init(&at_lock, 0);
  if (icv->stacksize != -1) marcel_attr_setstacksize(&mattr, icv->stacksize);
  threadlimit = icv->threadlimit;
  return (0);
}


void othr_finalize(int exitvalue)
{                  
}


/* The function executed by each thread */
void *threadjob(void *env)
{
  thr_info_t *thri = (thr_info_t *)env;

  ort_ee_dowork((int) thri->thr_id, thri->team_argument);  /* Execute code */  
  return 0;
}


/* Request for "numthr" threads to execute parallelism in level "level" */
int othr_request(int numthr, int level)
{
  int res, tmp;
  int omp_get_dynamic();
 
  if ((threadlimit == -1)||(!omp_get_dynamic())) {
        return (numthr);
  }

  marcel_mutex_lock(&at_lock);
  tmp = activethreads + numthr;
  if (tmp <= threadlimit) {
        activethreads = tmp;
  }
  else {
        numthr = threadlimit - activethreads;
        activethreads = threadlimit;
  }
  marcel_mutex_lock(&at_lock);

  return (numthr);
}

/* Dispatches numthreads from the pool-list and gives them work to do */
void othr_create(int numthr, int level, void *arg, void **info)
{
  int i;
  struct team_info *ti;

  if (numthr <= 0) return;
  if (*info == NULL) {
    ti = calloc(1, sizeof(struct team_info));
  } 
  else {
    ti = *info;
  }

  if (ti->team_thr == NULL) {
    ti->team_thr = calloc(1, (numthr+1)*sizeof(thr_info_t));	
  }
  else {
    if (ti->team_nthr < (numthr+1)) {
      ti->team_thr = realloc(ti->team_thr, (numthr+1)*sizeof(thr_info_t));
      memset(ti->team_thr, 0, (numthr+1)*sizeof(thr_info_t));
    }
  }

  ti->team_nthr = numthr + 1;
  *info = (void *)ti;

  for (i = 1; i <= numthr; i++) {
    ti->team_thr[i].team_argument = arg;
    ti->team_thr[i].thr_id = i;
    marcel_create(&ti->team_thr[i].thr, &mattr, threadjob, 
                  (void*) &ti->team_thr[i]);
  }

}


/* Only the master thread can call this.
 * It blocks the thread waiting for all its children to finish their job.
 */
void othr_waitall(void **info)
{
  struct team_info *ti = (struct team_info *) (*info);
  int i;

  for (i = 1; i <= (ti->team_nthr-1); i++)
      marcel_join(ti->team_thr[i].thr, NULL);

  if ((threadlimit != -1)|| omp_get_dynamic()) {
    marcel_mutex_lock(&at_lock);
    activethreads -= ti->team_nthr;
    marcel_mutex_unlock(&at_lock);
  }
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                 *
 *  LOCKS                                                          *
 *                                                                 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


int othr_init_lock(othr_lock_t *lock, int type)
{
  switch (lock->lock.type = type)
  {                                               
    case ORT_LOCK_NEST:
    {
      othr_nestlock_t *l = &(lock->lock.data.nest);

      marcel_mutex_init(&l->ilock, 0);
      marcel_mutex_init(&l->lock, 0);
      l->count = 0;    
      marcel_cond_init(&l->cond, 0);
      return (0);
    }

    case ORT_LOCK_SPIN:        
    {
      lock->lock.data.spin.rndelay = 0;
      return marcel_mutex_init(&(lock->lock.data.spin.mutex), 0);
    }

    default: /* ORT_LOCK_NORMAL */
      return marcel_mutex_init(&(lock->lock.data.normal), 0);
  }
}

int othr_destroy_lock(othr_lock_t *lock)
{
  switch (lock->lock.type)
  {
    case ORT_LOCK_NEST:
    {
      othr_nestlock_t *l = &(lock->lock.data.nest);

      marcel_mutex_destroy(&l->lock);
      marcel_cond_destroy(&l->cond);
      marcel_mutex_destroy(&l->ilock);
      return 0;
    }

    case ORT_LOCK_SPIN:
      return marcel_mutex_destroy(&(lock->lock.data.spin.mutex));

    default: /* ORT_LOCK_NORMAL */
      return marcel_mutex_destroy(&(lock->lock.data.normal));
  }
}

int othr_set_lock(othr_lock_t *lock)
{
  switch (lock->lock.type)
  {
    case ORT_LOCK_NEST:
    {
      othr_nestlock_t *l = &(lock->lock.data.nest);
      marcel_t me = marcel_self();

      marcel_mutex_lock(&l->ilock);
        if (marcel_mutex_trylock(&l->lock)) /* If not locked, lock it */
        {
          l->owner = me;              /* Get ownership */
          l->count++;
        }
        else
          if (l->owner == me)  /* Did i do it? */
            l->count++;
          else                                    /* Locked by someone else */
          {
            while ( !marcel_mutex_trylock(&l->lock) )
              marcel_cond_wait(&l->cond, &l->ilock);
            l->owner = me;
            l->count++;
          }
      marcel_mutex_unlock(&l->ilock);
      return (0);
    }

    case ORT_LOCK_SPIN:
    {
      /* General, portable solution: spin trying with exponential backoff */
      volatile int count, delay, dummy;
      for (delay = lock->lock.data.spin.rndelay;
                   !marcel_mutex_trylock(&(lock->lock.data.spin.mutex)); )
      {
        for (count = dummy = 0; count < delay; count++ )
          dummy += count;      /* To avoid compiler optimizations */
        if (delay == 0)
          delay = 1;
        else
          if (delay < 10000)     /* Don't delay too much */
            delay = delay << 1;
      }
      lock->lock.data.spin.rndelay++;    /* Next thread would wait a bit more */
      return (0);
    }

    default: /* ORT_LOCK_NORMAL */
      return marcel_mutex_lock(&(lock->lock.data.normal));
  }
}

int othr_unset_lock(othr_lock_t *lock)
{
  switch (lock->lock.type)
  {
    case ORT_LOCK_NEST:
    {
      othr_nestlock_t *l = &(lock->lock.data.nest);
      marcel_t me = marcel_self();

      marcel_mutex_lock(&l->ilock);
        if ((l->owner == me) && (l->count > 0))
        {
          l->count--;
          if (l->count == 0)
          {
            marcel_mutex_unlock(&l->lock);
            marcel_cond_signal(&l->cond);
          }
        }
      marcel_mutex_unlock(&l->ilock);
      return 0;                
    }

    case ORT_LOCK_SPIN: 
      lock->lock.data.spin.rndelay = 0;   /* reset it */
      return marcel_mutex_unlock(&(lock->lock.data.spin.mutex));

    default: /* ORT_LOCK_NORMAL */         
      return marcel_mutex_unlock(&(lock->lock.data.normal));
  }      
}


int othr_test_lock(othr_lock_t *lock)
{
  switch (lock->lock.type)
  {
    case ORT_LOCK_NEST:
    {
      othr_nestlock_t *l = &(lock->lock.data.nest);
      int res;
      marcel_t me = marcel_self();

      marcel_mutex_lock(&l->ilock);
        if (marcel_mutex_trylock(&l->lock))
        {
          l->owner = me;
          res = ++l->count;
        }
        else
          if (l->owner == me)
            res = ++l->count;
          else
            res = 0;
      marcel_mutex_unlock(&l->ilock);
      return res;
    }

    case ORT_LOCK_SPIN:
      return ( marcel_mutex_trylock(&(lock->lock.data.spin.mutex)) );

    default: /* ORT_LOCK_NORMAL */
      return ( marcel_mutex_trylock(&(lock->lock.data.normal)) );
  }
}

extern void __ompi_main(int argc, char *argv[]);

int marcel_main(int argc, char *argv[])
{
  marcel_init(&argc, argv);
  __ompi_main(argc, argv);

  return 0;
}
