/*
  OMPi OpenMP Compiler
  == Copyright since 2001 the OMPi Team
  == Department of Computer Science, University of Ioannina

  This file is part of OMPi.

  OMPi is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  OMPi is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with OMPi; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* 
 * 16 Dec. 2009
 *   OpenMP 3.0 additions
 * 3 Nov. 2006
 *   Modified othr_create() and moved here all the pool-related stuff.
 * 31 Oct 2006
 *   Added othr_yield()
 */
 
/*
 * An implementation of the othr API using plain & portable POSIX threads.
 * A simple, thread library for single-level parallelism.
 */

#include "../ort.h"
#include "ee.h"

#define WAIT_WHILE(f) { FENCE; for(;f;) othr_yield(); }

static int pthread_lib_inited = 0;     /* Flag to avoid re-initializations */
static ptrint num_created;             /* How many threads exist in total */
static volatile int running = 0;       /* # threads busy */
static othr_lock_t runlock;            /* Lock for accessing "running" */
static volatile int spin = 1;          /* Threads spin here waiting for work */
static volatile aligned_int *privatespin,     /* Each thread waits ... */
                            *actualarray;

/* Global stuff for the team */
static void *team_argument;            /* The argument each thread must init */
static volatile int team_nthr;         /* # threads in the current team */
static pthread_attr_t *globalattr;     /* Specific stacksize */

/* variables from ort */
static int stacksize, threadlimit;


static 
void _create_bunch_of_threads(int numthr)
{
  pthread_t thr;
  int e;
  extern void *threadjob(void *);

  for (; num_created < numthr; )
  {
    privatespin[++num_created].value = 1;
    FENCE;
    if ( (e = pthread_create(&thr, globalattr, threadjob, (void*)num_created)) )
      ort_error(5, "pthread_create() failed with %d\n", e);
  }
}


/* Library initialization.
 * Just create an intial pool of threads.
 */
int othr_initialize(int *argc, char ***argv, ort_icvs_t *icv, ort_caps_t *caps)
{
  int nthr;

  stacksize   = icv->stacksize;    /* New thread's stacksize  */
  threadlimit = icv->threadlimit;  /* Max active threads */

  nthr = (icv->nthreads >= 0) ?    /* Explicitely requested population */
               icv->nthreads :
               icv->ncpus - 1;     /* we don't handle the initial thread */
  if (threadlimit > 0 && nthr > threadlimit)
    nthr = threadlimit;

  caps->supports_nested            = 0;
  caps->supports_dynamic           = 1;
  caps->supports_nested_nondynamic = 0;
  caps->max_levels_supported       = 1;
  caps->default_numthreads         = nthr;
  caps->max_threads_supported      = -1;     /* No limit */

  num_created = 0;
  if ( pthread_lib_inited ) return (0);
  
  globalattr = NULL;
  if (stacksize > 0)
  {
    if (pthread_attr_init(globalattr)) 
      ort_error(5, "pthread_attr_init() failed\n");    
    if (pthread_attr_setstacksize(globalattr, stacksize)) 
      ort_error(5, "pthread_attr_setstacksize() failed\n");
  }

  if (nthr > 0)
  {
  #ifdef HAVE_PTHREAD_SETCONCURRENCY
    pthread_setconcurrency(nthr+1);
  #endif

    privatespin = (aligned_int *)    /* private spinners */
        ort_alloc_aligned((nthr+1)*sizeof(aligned_int), (void **) &actualarray);
    othr_init_lock(&runlock, ORT_LOCK_SPIN);

    _create_bunch_of_threads(nthr);  /* Create pool of threads */
  }

  num_created = nthr;
  pthread_lib_inited = 1;
  return (0);
}


void othr_finalize(int exitvalue)
{
}


/* The function executed by each thread.
 * The threads wait on "spin" *AND* on "myid > team_nthr".
 * The latter is required so that ONLY num_nthr threads
 * wake up when "spin" becomes 0.
 */
void *threadjob(void *id)
{
  int myid = (int)((ptrint) id);
  
  for (;;)
  {
    /* Wait for work */
    WAIT_WHILE(spin || myid > team_nthr);

    if (ORT_DEBUG) ort_debug_thread("about to call ort_ee_work()");
    ort_ee_dowork(myid, team_argument);      /* Execute requested code */

#if defined(HAVE_ATOMIC_FAA) && !defined(EE_TYPE_PROCESS)
    _faa(&(running), -1);
#else
    othr_set_lock(&runlock);
      running--;                      /* Decrement number of running threads */
    othr_unset_lock(&runlock);
#endif
    FENCE;

    /* Spin till master releases me */
    WAIT_WHILE(privatespin[myid].value);
    privatespin[myid].value = 1;                   /* Prepare for next round */
  }
}


/* We only support 1 level of parallelism, so we always return 0 if level > 1.
 */
int othr_request(int numthr, int level)
{
  if (threadlimit > 0 && numthr > threadlimit)
    numthr = threadlimit;
  return ( (level == 1) ? numthr : 0 );
}


/* Dispatches numthreads from the pool and gives them work to do;
 * if the pool is insufficient, new threads will be created from scratch.
 * Here we assume that all threads are spinning around the global
 * spin (not their own private spins).
 */
void othr_create(int numthr, int level, void *arg, void **ignore)
{
  int ret, e;

  if (numthr <= 0) return;

  if (threadlimit > 0 && numthr > threadlimit)
    numthr = threadlimit;

  if (numthr > num_created)
  {
    privatespin = (aligned_int *)   /* Resize the private spin array */
                  ort_realloc_aligned((numthr+32)*sizeof(aligned_int),
                                      (void **) &actualarray);
    FENCE;
    _create_bunch_of_threads(numthr);
  }

  team_argument = arg;
  team_nthr     = running = numthr;
  FENCE;

  spin = 0;       /* Release the threads */
}


/* Only the master thread can call this.
 * It blocks the thread waiting for all its children to finish their job.
 */
void othr_waitall(void **ignore)
{
  int i;

  WAIT_WHILE(running > 0);
  spin = 1;                              /* Block threads for next round */
  FENCE;
  for (i = 1; i <= team_nthr; i++)
    privatespin[i].value = 0;            /* Release all threads */
  FENCE;
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                 *
 *  LOCKS (identical to othr_pthreads)                             *
 *                                                                 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


int othr_init_lock(othr_lock_t *lock, int type)
{
  switch (lock->lock.type = type)
  {
    case ORT_LOCK_NEST:
    {
      othr_nestlock_t *l = &(lock->lock.data.nest);
      
      pthread_mutex_init(&l->ilock, NULL);
      pthread_mutex_init(&l->lock, NULL);
      l->count = 0;
      pthread_cond_init(&l->cond, 0);
      return (0);
    }
    
    case ORT_LOCK_SPIN:
    {
#ifdef HAVE_SPINLOCKS
      return pthread_spin_init(&(lock->lock.data.spin), 0);
#else
  #ifdef PTHREAD_MUTEX_SPINBLOCK_NP                           /* e.g. IRIX */
      static pthread_mutexattr_t spinblock_attr;
      static int firstCall = 1;
  
      if (firstCall) {
        firstCall = 0;
        pthread_mutexattr_init(&spinblock_attr);
        pthread_mutexattr_settype(&spinblock_attr, PTHREAD_MUTEX_SPINBLOCK_NP);
      }
      lock->lock.data.spin.rndelay = 0;
      return pthread_mutex_init(&(lock->lock.data.spin.mutex), &spinblock_attr);
  #else
      lock->lock.data.spin.rndelay = 0;
      return pthread_mutex_init(&(lock->lock.data.spin.mutex), NULL);
  #endif
#endif
    }
    
    default: /* ORT_LOCK_NORMAL */
      return pthread_mutex_init(&(lock->lock.data.normal), NULL);
  }    
}


int othr_destroy_lock(othr_lock_t *lock)
{
  switch (lock->lock.type)
  {
    case ORT_LOCK_NEST:
    {
      othr_nestlock_t *l = &(lock->lock.data.nest);
      
      pthread_mutex_destroy(&l->lock);
      pthread_cond_destroy(&l->cond);
      pthread_mutex_destroy(&l->ilock);
      return 0;
    }
      
    case ORT_LOCK_SPIN:
#ifdef HAVE_SPINLOCKS
      return pthread_spin_destroy(&(lock->lock.data.spin));
#else
      return pthread_mutex_destroy(&(lock->lock.data.spin.mutex));
#endif

    default: /* ORT_LOCK_NORMAL */
      return pthread_mutex_destroy(&(lock->lock.data.normal));
  }    
}


int othr_set_lock(othr_lock_t *lock)
{
  switch (lock->lock.type)
  {
    case ORT_LOCK_NEST:
    {
      othr_nestlock_t *l = &(lock->lock.data.nest);
      pthread_t       me = pthread_self();

      pthread_mutex_lock(&l->ilock);
        if (pthread_mutex_trylock(&l->lock) == 0) /* If not locked, lock it */
        {
          l->owner = me;                          /* Get wownership */
          l->count++;
        }
        else
          if (pthread_equal(l->owner, me))        /* Did i do it? */
            l->count++;
          else                                    /* Locked by someone else */
          {
            while ( pthread_mutex_trylock(&l->lock) )
              pthread_cond_wait(&l->cond, &l->ilock);
            l->owner = me;
            l->count++;
          }
      pthread_mutex_unlock(&l->ilock);
      return (0);
    }
    
    case ORT_LOCK_SPIN:
    {
#ifdef HAVE_SPINLOCKS
      return pthread_spin_lock(&(lock->lock.data.spin));
#else
  #ifdef PTHREAD_MUTEX_SPINBLOCK_NP                             /* e.g. IRIX */
      return pthread_mutex_lock(&lock->lock.data.spin.mutex);
  #else
      /* General, portable solution: spin trying with exponential backoff */
      volatile int count, delay, dummy;
      for (delay = lock->lock.data.spin.rndelay;
                   pthread_mutex_trylock(&(lock->lock.data.spin.mutex)); )
      {
        for (count = dummy = 0; count < delay; count++ )
          dummy += count;      /* To avoid compiler optimizations */
        if (delay == 0)        
          delay = 1;
        else
          if (delay < 10000)     /* Don't delay too much */
            delay = delay << 1;
      }
      lock->lock.data.spin.rndelay++;    /* Next thread would wait a bit more */
      return (0);
  #endif
#endif
    }
    
    default: /* ORT_LOCK_NORMAL */
      return pthread_mutex_lock(&(lock->lock.data.normal));
  }    
}


int othr_unset_lock(othr_lock_t *lock)
{
  switch (lock->lock.type)
  {
    case ORT_LOCK_NEST:
    {
      othr_nestlock_t *l = &(lock->lock.data.nest);
      
      pthread_mutex_lock(&l->ilock);
        if (pthread_equal(l->owner, pthread_self()) && l->count > 0)
        {
          l->count--;
          if (l->count == 0)
          {
            pthread_mutex_unlock(&l->lock);
            pthread_cond_signal(&l->cond);
          }
        }
      pthread_mutex_unlock(&l->ilock);
      return 0;
    }
     
    case ORT_LOCK_SPIN:
#ifdef HAVE_SPINLOCKS
      return pthread_spin_unlock(&(lock->lock.data.spin));
#else
      lock->lock.data.spin.rndelay = 0;   /* reset it */
      return pthread_mutex_unlock(&(lock->lock.data.spin.mutex));
#endif

    default: /* ORT_LOCK_NORMAL */
      return pthread_mutex_unlock(&(lock->lock.data.normal));
  }    
}


int othr_test_lock(othr_lock_t *lock)
{
  switch (lock->lock.type)
  {
    case ORT_LOCK_NEST:
    {
      othr_nestlock_t *l = &(lock->lock.data.nest);
      int res;

      pthread_mutex_lock(&l->ilock);
        if (pthread_mutex_trylock(&l->lock) == 0)
        {
          l->owner = pthread_self();
          res = ++l->count;
        }
        else
          if (pthread_equal(l->owner, pthread_self()))
            res = ++l->count;
          else
            res = 0;
      pthread_mutex_unlock(&l->ilock);
      return res;
    }
     
    case ORT_LOCK_SPIN:
#ifdef HAVE_SPINLOCKS
      return pthread_spin_trylock(&(lock->lock.data.spin));
#else
      return ( !pthread_mutex_trylock(&(lock->lock.data.spin.mutex)) );
#endif

    default: /* ORT_LOCK_NORMAL */
      return ( !pthread_mutex_trylock(&(lock->lock.data.normal)) );
  }    
}
