/*
  OMPi OpenMP Compiler
  == Copyright since 2001 the OMPi Team
  == Department of Computer Science, University of Ioannina

  This file is part of OMPi.

  OMPi is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  OMPi is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with OMPi; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* 
 * Dec. 2009
 *   OpenMP 3.0 additions & suuport for dynamic pool (EL)
 * July 2007
 *   Bug fix (thanks to Marie-Christine Counilh)
 * May 2007
 *   Support for nested parallelism by George Filos
 */

#include "../ort.h"
#include <stdlib.h>
#include <stdio.h>
#include "ee.h"

#define WAIT_WHILE(f) { FENCE; for(;f;) othr_yield(); }

/* Types & declarations
 */
typedef struct {
  othr_lock_t  rlock;
  volatile int running;         /* # running threads in a team */
} othr_info;

typedef struct othr_pool_s {
  void *arg;                    /* The function's argument */
  int id;                       /* A sequential id given by OMPI */
  void *info;
  int spin;                     /* Spin here waiting for work */
  struct  othr_pool_s *next;    /* Next node into the list */
} othr_pool_t;

static int   solthr_lib_inited = 0;      /* Flag to avoid re-initializations */
static othr_pool_t  *H = NULL;                  /* Pool = a list of threads */
static volatile int plen;                          /* # threads in the pool */
static othr_lock_t  plock;                  /* A lock for accesing the pool */

/* OpenMP 3.0 limits */
static int threadlimit, threadscreated, levellimit;
static int stacksize = -1;      /* For stacksize */


/* A list of blocked threads */
static
othr_pool_t *_new_bunch_of_threads(int n, othr_pool_t **tail)
{
  pthread_t   thr;
  othr_pool_t *block, *node;
  int         i, e;
  void        *threadjob(void *);

  if ((block=(othr_pool_t*)ort_alloc_aligned(n*sizeof(othr_pool_t),NULL))==NULL)
    ort_error(5, "othr_init() failed; could not allocate memory\n");
  for (i = 0, node = block; i < n; i++)
  {
    node->spin = 1; FENCE;
    node->next = (node+1);
    if ( (e = thr_create(NULL,NULL,threadjob, (void*)node,THR_BOUND,&thr)) )
      ort_error(5, "pthread_create() failed with %d\n", e);
    node++;
  }
  (--node)->next = NULL;
  if (tail) *tail = node;
  return (block);
}


/* Library initialization
 */
int othr_initialize(int *argc, char ***argv, ort_icvs_t *icv, ort_caps_t *caps)
{
  int nthr;

  nthr = (icv->nthreads >= 0) ?  /* Explicitely requested population */
               icv->nthreads :
               icv->ncpus - 1;   /* we don't handle the initial thread */
  if (icv->threadlimit != -1 && nthr > icv->threadlimit)
    nthr = icv->threadlimit;

  caps->supports_nested            = 1;
  caps->supports_dynamic           = 1;
  caps->supports_nested_nondynamic = 1;
  caps->max_levels_supported       = -1;     /* No limit */
  caps->default_numthreads         = nthr;
  caps->max_threads_supported      = -1;     /*no limit*/
  
  if ( solthr_lib_inited ) return (0);
  
  threadlimit    = icv->threadlimit;
  threadscreated = 0;
  levellimit     = icv->levellimit;
  stacksize      = icv->stacksize ;
  if (nthr > 0)
  {
    thr_setconcurrency(nthr+1);

    H = _new_bunch_of_threads(nthr, NULL);      /* Create the initial pool */
    threadscreated = nthr;
    othr_init_lock(&plock, ORT_LOCK_SPIN);
  }

  plen = nthr;
  solthr_lib_inited = 1;
  return (0);
}


void othr_finalize(int exitvalue)
{
}


/* The function executed by each thread */
void *threadjob(void *env)
{
  othr_info *myinfo;
  othr_pool_t *env_t = (othr_pool_t *)env;
   
  while (1) {
    /* Wait for work */
    WAIT_WHILE(env_t->spin);

    env_t->spin = 1;                  /* Prepare me for next round */
    if (ORT_DEBUG) ort_debug_thread("about to call ort_ee_work()");
    ort_ee_dowork(env_t->id, env_t->arg);     /* Execute requested code */

    myinfo = env_t->info;             /* Moved this up - thanks to M-CC */
    othr_set_lock(&plock);
      env_t->next = H;                /* Re-enter the pool */
      H = env_t;
      plen++;
    othr_unset_lock(&plock);
    FENCE;

    /* Update the "running" field of my parent's node */
    othr_set_lock(&myinfo->rlock);
      myinfo->running--;           /* Running threads are one less in my team */
    othr_unset_lock(&myinfo->rlock);
    FENCE;
  }
}


/* Request for "numthr" threads to execute parallelism in level "level" */
int othr_request(int numthr, int level)
{
  int         new, tmpplen;
  othr_pool_t *bunch, *tail;

  if (level == 1)      /* Only the master thread is here, no need for locking */
  {
    if (numthr <= plen)
      plen -= numthr;
    else
    {
      new = numthr - plen;
      if (threadlimit != -1 && new+threadscreated > threadlimit)
        new = threadlimit - threadscreated;

      bunch = _new_bunch_of_threads(new, &tail);
      tail->next = H;
      H = bunch;

      threadscreated += new;
      numthr = plen + new;
      plen=0;
    }
		
  }
  else
  {
    if (levellimit != -1 && level > levellimit)   
      return 0;
    othr_set_lock(&plock);  /* lock in order to check the limits */
    if (numthr <= plen)
    {
      plen -= numthr;
      othr_unset_lock(&plock);
      FENCE;
    }
    else
    {
      new = numthr - plen;
      if (threadlimit != -1 && new+threadscreated > threadlimit)
        new = threadlimit - threadscreated;
	
      threadscreated += new;
      tmpplen = plen;
      plen = 0;
      othr_unset_lock(&plock);

      if (new == 0) return (tmpplen);
	
      bunch = _new_bunch_of_threads(new, &tail); /* Augment the pool */
      othr_set_lock(&plock);
        tail->next = H;
        H = bunch;
        FENCE;
      othr_unset_lock(&plock);
      numthr = tmpplen + new;
    }
  }
  return (numthr);
}


/* Dispatches numthreads from the pool and gives them work to do */
void othr_create(int numthr, int level, void *arg, void **info)
{
  othr_pool_t *p;
  othr_info   *t = (othr_info *) *info;
  int         i;
 
  if (t == NULL)  /* Thread becomes a parent for the first time */
  {
    t = (othr_info *) ort_alloc_aligned(sizeof(othr_info),NULL);
    othr_init_lock(&t->rlock, ORT_LOCK_SPIN);
  }
  t->running = numthr;
 
  /* Wake up "nthr" threads and give them work to do */
  for (i = 1; i <= numthr; i++)
  {
    othr_set_lock(&plock);
      p = H;
      H = H->next;
    othr_unset_lock(&plock);
     
    p->arg  = arg;
    p->info = t;
    p->id   = i;
    FENCE;
    p->spin = 0;   /* Release thread i */
  }
  *info = t;
}


/* Only the master thread can call this.
 * It blocks the thread waiting for all its children to finish their job.
 */
void othr_waitall(void **info)
{    
  othr_info *t = *info;
  
  WAIT_WHILE(t->running > 0);
}


void *othr_getspecific(othr_key_t key)
{
  void *result;

  thr_getspecific(key, &result);
  return ( result );
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                 *
 *  LOCKS (identical to othr_solthr1)                              *
 *                                                                 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


int othr_init_lock(othr_lock_t *lock, int type)
{
  switch (lock->lock.type = type)
  {
    case ORT_LOCK_NEST:
    {
      othr_nestlock_t *l = &(lock->lock.data.nest);
      
      mutex_init(&l->ilock, USYNC_THREAD, NULL);
      mutex_init(&l->lock, USYNC_THREAD, NULL);
      l->count = 0;
      cond_init(&l->cond, USYNC_THREAD, NULL);
      return (0);
    }
    
    case ORT_LOCK_SPIN:
      lock->lock.data.spin.rndelay = 0;
      return mutex_init(&(lock->lock.data.spin.mutex), USYNC_THREAD, NULL);
    
    default: /* ORT_LOCK_NORMAL */
      return mutex_init(&(lock->lock.data.normal), USYNC_THREAD, NULL);
  }    
}


int othr_destroy_lock(othr_lock_t *lock)
{
  switch (lock->lock.type)
  {
    case ORT_LOCK_NEST:
    {
      othr_nestlock_t *l = &(lock->lock.data.nest);
      
      mutex_destroy(&l->lock);
      mutex_destroy(&l->ilock);
      cond_destroy(&l->cond);
      return 0;
    }
      
    case ORT_LOCK_SPIN:
      return mutex_destroy(&(lock->lock.data.spin.mutex));

    default: /* ORT_LOCK_NORMAL */
      return mutex_destroy(&(lock->lock.data.normal));
  }    
}


int othr_set_lock(othr_lock_t *lock)
{
  switch (lock->lock.type)
  {
    case ORT_LOCK_NEST:
    {
      othr_nestlock_t *l = &(lock->lock.data.nest);
      thread_t        me = thr_self();

      mutex_lock(&l->ilock);
        if (mutex_trylock(&l->lock) == 0) /* If not locked, lock it */
        {
          l->owner = me;                  /* Get wownership */
          l->count++;
        }
        else
          if (l->owner == me)             /* Did i do it? */
            l->count++;
          else                            /* Locked by someone else */
          {
            while ( mutex_trylock(&l->lock) )
              cond_wait(&l->cond, &l->ilock);
            l->owner = me;
            l->count++;
          }
      mutex_unlock(&l->ilock);
      return (0);
    }
    
    case ORT_LOCK_SPIN:
    {
      /* General, portable solution: spin trying with exponential backoff */
      volatile int count, delay, dummy;
      for (delay = lock->lock.data.spin.rndelay;
                   mutex_trylock(&(lock->lock.data.spin.mutex)); )
      {
        for (count = dummy = 0; count < delay; count++ )
          dummy += count;      /* To avoid compiler optimizations */
        if (delay == 0)        
          delay = 1;
        else
          if (delay < 10000)     /* Don't delay too much */
            delay = delay << 1;
      }
      lock->lock.data.spin.rndelay++;    /* Next thread would wait a bit more */
      return (0);
    }
    
    default: /* ORT_LOCK_NORMAL */
      return mutex_lock(&(lock->lock.data.normal));
  }    
}


int othr_unset_lock(othr_lock_t *lock)
{
  switch (lock->lock.type)
  {
    case ORT_LOCK_NEST:
    {
      othr_nestlock_t *l = &(lock->lock.data.nest);
      
      mutex_lock(&l->ilock);
        if (l->owner == thr_self() && l->count > 0)
        {
          l->count--;
          if (l->count == 0)
          {
            mutex_unlock(&l->lock);
            cond_signal(&l->cond);
          }
        }
      mutex_unlock(&l->ilock);
      return 0;
    }
     
    case ORT_LOCK_SPIN:
      lock->lock.data.spin.rndelay = 0;   /* reset it */
      return mutex_unlock(&(lock->lock.data.spin.mutex));

    default: /* ORT_LOCK_NORMAL */
      return mutex_unlock(&(lock->lock.data.normal));
  }    
}


int othr_test_lock(othr_lock_t *lock)
{
  switch (lock->lock.type)
  {
    case ORT_LOCK_NEST:
    {
      othr_nestlock_t *l = &(lock->lock.data.nest);
      int res;

      mutex_lock(&l->ilock);
        if (mutex_trylock(&l->lock) == 0)
        {
          l->owner = thr_self();
          res = ++l->count;
        }
        else
          if (l->owner == thr_self())
            res = ++l->count;
          else
            res = 0;
      mutex_unlock(&l->ilock);
      return res;
    }
     
    case ORT_LOCK_SPIN:
      return ( !mutex_trylock(&(lock->lock.data.spin.mutex)) );

    default: /* ORT_LOCK_NORMAL */
      return ( !mutex_trylock(&(lock->lock.data.normal)) );
  }    
}
