/*
  OMPi OpenMP Compiler
  == Copyright since 2001 the OMPi Team
  == Department of Computer Science, University of Ioannina

  This file is part of OMPi.

  OMPi is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  OMPi is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with OMPi; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* ort.c -- OMPi RunTime library */

/*
 * Version 1.0.2b
 *   ort_ee_dowork() (old ort_get_thread_work()) now also call the thread
 *   function.
 * Version 1.0.1j
 *   ort.c code split & reorganization; see ort_*.c
 * Version 1.0.0e
 *   added OpenMP 3.0 runtime functions
 * Version 1.0.0a
 *   added atomic operations
 * Version 0.9.9
 *   thrinfo renamed to eecb; thrinfo block recyler ditched;
 *   ort_get_thread_work() changed.
 * Version 0.9.2
 *   Recycling of thrinfo team structs; added level to ee_create().
 * Version 0.9.1
 *   New & better threadprivate code
 *   Ditched some functions (ort_destroy_team, ort_assign_key, etc).
 *   ort_create_team() --> ort_execute_parallel().
 * Version 0.8.6
 *   More bug fixes; fixed support for both clock_gettime & gettimeofday.
 * Version 0.8.5n.x
 *   Small fixes and improvements / added ort_finalize().
 * Version 0.8.4.8
 *   19 Nov. 2006
 *     Too many changes to mention. Added code for critical & atomic,
 *     moved stuff around, made cache-aligned allocators, move omp.c code
 *     here, ...
 * Version 0.8.4.5
 *   3 Nov. 2006
 *     Removed all pool-related stuff and moved it to the thread library
 *   5 Nov. 2006
 *     Fixed copyprivate parser bug & added runtime copyprivate support.
 * Version 0.8.4.4
 *   Removed locks for threads waiting for work; now they spin.
 * Version 0.8.4.3
 *   31 Oct. 2006
 *     New code for ORDERED (uses busy waiting and no strange locking.
 *     Also removed the join_lock (which was also locked/unlocked in
 *          non-standard manner). Join is also based on bbusy-waiting.
 *     Now TT structrures are mostly initialized by each thread, not
 *          by the master thread.
 *     _destroy_team() now takes no argument.
 * Version 0.8.4
 *   Too many changes: SSF rounds gone for ever; schedule support
 *   revamped; ordered improved; num_single/sections/for gone away;
 *   and many many others.
 * Version 0.8.3
 *   (1) For SECTIONS and SINGLE, SSF rounds are not used any more.
 *       Now they use a completely different system, through 
 *       enter_workshare_region()
 *   (2) sched_info removed completely; It was used in every SSF round
 *       for remembering the runtime schedule of FORs. Now all enviromental
 *       variables are read once in the beginning (_omp_initialize())
 *       and they are saved in global variables.
 */
 
#include "ort_prive.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <assert.h>


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * GLOBAL VARIABLES / DEFINITIONS / MACROS                           *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
                                     

/* All global variables ORT handles; if ee=proc, this is also placed
 * in shared memory.
 */
static ort_vars_t ort_globals;  /* All ORT globals stored here */
ort_vars_t        *ort;         /* Pointer to ort_globals */

static int        ort_initialized = 0;

/* ort_master is *only* needed for threadprivate variables (nothing else) */
static ort_eecb_t      *ort_master;  /* eecb for the initial thread */
static ort_task_node_t *master_task; /* task structure for the initial task */

/* Execution entity (thread/process) control block */
#ifdef USE_TLS
  TLS_KEYWORD void *myeecb;
#else
  ee_key_t eecb_key;  /* Key for ort's "thread"-specific data; the actual
                         data is a pointer to the eecb */
#endif

/* Handy macro */
#define initialize_eecb(eecb) {\
     (eecb)->parent            = NULL;\
     (eecb)->sdn               = (eecb);\
     (eecb)->num_children      = 0;\
     (eecb)->num_siblings      = 1;     /* We are just 1 thread! */\
     (eecb)->thread_num        = 0;\
     (eecb)->level             = 0;     /* The only one in level 0 */\
     (eecb)->activelevel       = 0;     /* The only one in level 0 */\
     (eecb)->shared_data       = 0;\
     (eecb)->mynextNWregion    = 0;\
     (eecb)->have_created_team = 0;\
     (eecb)->ee_info           = NULL;  /* *Must* init to NULL */\
   }

/* Process id */
#if defined(EE_TYPE_PROCESS)
  #define __MYPID  ee_pid()
#else
  #define __MYPID  0
#endif


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * INITIALIZATION / SHUTDOWN                                         *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/*
 * Last function called (just before exit).
 * "exitval" is what the original (user) main() returned.
 */
void ort_finalize(int exitval)
{
  ee_finalize(exitval);
  _exit(exitval);             /* Make sure nothing else is called */
}


/* Called only if the user calls exit() */
void _at_exit_call(void)
{
  ee_finalize(-1);
}


/* size | sizeB | sizeK | sizeM | sizeG */
static long int _size_bytes(char *str)
{
  long int size;
  char *s, mod;

  for (s = str; isspace(*s) || isdigit(*s); s++)
    ;
  switch (*s)  /* specifier */
  {
    case   0: 
    case 'k': case 'K': mod = 10; break;  /* default */
    case 'b': case 'B': mod = 0;  break;
    case 'm': case 'M': mod = 20; break;
    case 'g': case 'G': mod = 30; break;
    default:
      goto WRONG_STSIZE;
  };
  if (*s)      /* past the specifier */
  {
    for (*s = 0, s++; isspace(*s); s++) /* */ ;
    if (*s) goto WRONG_STSIZE;
  }

  if ((size = atoi(str)) < 0) goto WRONG_STSIZE;
  if (mod) size <<= mod;
  if (size < 128) 
  {
    ort_warning("illegal OMP_STACKSIZE value (%ld); using 256K instead\n",size);
    size = KBytes(256);
  }
  return (size);

  WRONG_STSIZE:
    ort_warning("incorrect OMP_STACKSIZE specification; using 256K.\n", str);
    return ( KBytes(256) );
}


#define ISTRUE(s) \
          (strcmp(s,"true") == 0 || strcmp(s,"TRUE") == 0 || strcmp(s,"1") == 0)
          

/* Environmental variables */
void ort_get_environment()
{
  char *s, *t;
  int  n;

  if ((s = getenv("OMP_NUM_THREADS")) != NULL)
    if (sscanf(s, "%d", &n) == 1 && n > 0)
      ort->icvs.nthreads = n;
  
  if ((s = getenv("OMP_DYNAMIC")) != NULL)
  {
    if (ISTRUE(s))
      ort->icvs.dynamic = 1;
    else
      ort->icvs.dynamic = 0;
  }

  if ((s = getenv("OMP_NESTED")) != NULL)
  {
    if (ISTRUE(s))
      ort->icvs.nested = 1;
    else
      ort->icvs.nested = 0;
  }
  
  if ((s = getenv("OMP_SCHEDULE")) != NULL)
  {
    if ((t = strchr(s, ',')) != NULL)   /* Chunksize specified. */
    {
      sscanf(t+1, "%d", &n);
      if (n > 0)
        ort->icvs.rtchunk = n;
      else
        ort_warning("illegal chunksize in environment "
                    "variable OMP_SCHEDULE; using default value\n");
    }
    if (strncmp(s, "static", 6) == 0)
      ort->icvs.rtschedule = omp_sched_static;
    else if (strncmp(s, "dynamic", 7) == 0)
      ort->icvs.rtschedule = omp_sched_dynamic;
    else if (strncmp(s, "guided", 6) == 0)
      ort->icvs.rtschedule = omp_sched_guided;
    else if (strncmp(s, "auto", 4) == 0)
      ort->icvs.rtschedule = omp_sched_auto;
    else
    {
      ort_warning("incorrect schedule type of environment "
                  "variable OMP_SCHEDULE; 'auto' assumed\n");
      ort->icvs.rtschedule = omp_sched_auto;
    }
  }

  /* OpenMP 3.0 */
  if ((s = getenv("OMP_STACKSIZE")) != NULL)
    ort->icvs.stacksize = _size_bytes(s);
  /* OpenMP 3.0 */
  if ((s = getenv("OMP_THREAD_LIMIT")) != NULL)
    if (sscanf(s, "%d", &n) == 1 && n > 0)
      ort->icvs.threadlimit = n;
  /* OpenMP 3.0 */
  if ((s = getenv("OMP_MAX_ACTIVE_LEVELS")) != NULL)
    if (sscanf(s, "%d", &n) == 1 && n > 0)
      ort->icvs.levellimit = n;
  /* OpenMP 3.0 */
  if ((s = getenv("OMP_WAIT_POLICY")) != NULL)
  {
    if (strncmp(s, "active", 6) == 0)
      ort->icvs.waitpolicy = _OMP_ACTIVE;
    else if (strncmp(s, "passive", 7) == 0)
      ort->icvs.waitpolicy = _OMP_PASSIVE;
    else
      ort_warning("incorrect value of environment "
                  "variable OMP_WAIT_POLICY; ignoring\n");
  }

}


/*
 * First function called.
 */
int ort_initialize(int *argc, char ***argv)
{
  if (ort_initialized) return 0;

  ort = &ort_globals;
  
  ort->icvs.ncpus       = ort_get_num_procs(); /* Default ICV values */
  ort->icvs.stacksize   = -1;                  /* OpenMP 3.0 */
  ort->icvs.threadlimit = -1;                  /* (unlimited) OpenMP 3.0 */
  ort->icvs.levellimit  = -1;                  /* (unlimited) OpenMP 3.0 */
  ort->icvs.waitpolicy  = _OMP_ACTIVE;         /* (ignored) OpenMP 3.0 */
  ort->icvs.nthreads    = -1;        /* per-task; no preference for now */
  ort->icvs.rtschedule  = omp_sched_auto;
  ort->icvs.rtchunk     = -1;
  ort->icvs.dynamic     = 1;
  ort->icvs.nested      = 0;

  ort_get_environment();                      /* Get environmental variables */

  /* Initialize the thread library */
  if (ort->icvs.nthreads > 0) ort->icvs.nthreads--; /* 1- for eelib */
  if ( ee_initialize(argc, argv, &ort->icvs, &ort->eecaps) != 0 )
    ort_error(1, "cannot initialize the thread library.\n");

  if (__MYPID == 0)               /* Everybody in case of threads */
  {
    /* Check for conformance to user requirements */
    if (ort->icvs.nthreads == -1)  /* Let the eelib set the default */
      ort->icvs.nthreads = ort->eecaps.default_numthreads+1;
    else                          /* user asked explicitely */
    {
      if (ort->eecaps.max_threads_supported > -1 &&
          ort->icvs.nthreads < ort->eecaps.max_threads_supported)
        if (!ort->icvs.dynamic || !ort->eecaps.supports_dynamic)
          ort_error(1, "the library cannot support the requested number (%d) "
                       "of threads.\n", ort->icvs.nthreads + 1);
      ort->icvs.nthreads++;        /* Restore value */
    }
    /* Fix discrepancies */
    if (ort->icvs.dynamic && !ort->eecaps.supports_dynamic)
      ort->icvs.dynamic = 0;
    if (ort->icvs.nested  && !ort->eecaps.supports_nested)
      ort->icvs.nested  = 0;
    check_nested_dynamic(ort->icvs.nested, ort->icvs.nested); /* is eelib ok? */

    /* OpenMP 3.0 stuff */
    if (ort->icvs.levellimit != -1)
      if (ort->eecaps.max_levels_supported != -1 &&
          ort->eecaps.max_levels_supported < ort->icvs.levellimit)
        ort->icvs.levellimit = ort->eecaps.max_levels_supported;
    if (ort->icvs.threadlimit != -1)
      if (ort->eecaps.max_threads_supported != -1 &&
          ort->eecaps.max_threads_supported < ort->icvs.threadlimit)
        ort->icvs.threadlimit = ort->eecaps.max_threads_supported;

    /* Initialize the 2 locks we need */
    ee_init_lock((ee_lock_t *) &ort->atomic_lock, ORT_LOCK_SPIN);
    ee_init_lock((ee_lock_t *) &ort->preparation_lock, ORT_LOCK_NORMAL);

    /* The "master" */
    ort_master = (ort_eecb_t *) ort_calloc_aligned(sizeof(ort_eecb_t), NULL);
    initialize_eecb(ort_master);
    ort_master->task_data = (ort_task_data_t *)
                            ort_calloc(sizeof(ort_task_data_t));

    /* The master's impicit task ("initial" task) */
    master_task = (ort_task_node_t *) ort_calloc(sizeof(ort_task_node_t));
    #if !defined(HAVE_ATOMIC_CAS)
      ee_init_lock((ee_lock_t*) &(master_task->lock),ORT_LOCK_NORMAL);
    #endif
    #if !defined(HAVE_ATOMIC_FAA)
      ee_init_lock((ee_lock_t*) &(ort_master->task_data->lock),ORT_LOCK_NORMAL);
    #endif
    master_task->icvs.dynamic    = ort->icvs.dynamic;
    master_task->icvs.nested     = ort->icvs.nested;
    master_task->icvs.rtschedule = ort->icvs.rtschedule;
    master_task->icvs.rtchunk    = ort->icvs.rtchunk;
    master_task->icvs.nthreads   = ort->icvs.nthreads;
    __CURRTASK(ort_master) = master_task;

#ifdef USE_TLS
#else
    ee_key_create(&eecb_key, 0);  /* This key stores a pointer to the eecb */
#endif
    __SETMYCB(ort_master);
    ort_init_tasking();
    if (ORT_DEBUG) ort_debug_thread("<this is the master thread>");

    /* Upon exit .. */
    atexit(_at_exit_call);
  }

#if defined(EE_TYPE_PROCESS)
  ort_share_globals();
#endif

  return ( ort_initialized = 1 );
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * SHARED MEMORY FOR THE PROCESS MODEL                               *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#if defined(EE_TYPE_PROCESS)

static ort_sglvar_t *ort_sglvar_list;   /* The list of requests */
static void         *ort_sglvar_area;   /* The start of the alloted memory */


void ort_share_globals()
{
  ort_sglvar_t *req;
  char         *mem;
  int          rsize;

  if (ort_sglvar_list != NULL)
    ort_shmalloc(&ort_sglvar_area, ort_sglvar_list->size, 1);

  for (mem = ort_sglvar_area, req = ort_sglvar_list; req != NULL; )
  {
    rsize = (req->next != NULL) ? req->size - req->next->size
                                : req->size;
    *(req->varptr) = (void *) mem;
    if (req->initvalue)
      memcpy(mem, req->initvalue, rsize);
    else
      memset(mem, 0, rsize);

    mem += rsize;

    req = (ort_sglvar_list = req)->next;
    free(ort_sglvar_list);
  }
}


/* Should mark all allocation requests; it is guaranteed by the parser
 * that this is called *before* main() starts. Thus, it should only file
 * the requests and do the actual allocations later, when ort_initialize()
 * is called (*dataptr should then be made to point to an allocated
 * space of size bytes).
 */
void ort_sglvar_allocate(void **varptr, int size, void *initer)
{
  ort_sglvar_t *req = (ort_sglvar_t *) ort_calloc(sizeof(ort_sglvar_t));

  req->varptr     = varptr;
  req->size       = ort_sglvar_list ? size + ort_sglvar_list->size : size;
  req->initvalue  = initer;
  req->next       = ort_sglvar_list;
  ort_sglvar_list = req;
}

#endif


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * THREADS & THEIR TEAMS                                             *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
                                     
                                     
/* Here an ee initializes its control block and calls the actual thread 
 * function (this should be called from the ee library, by the actual ee).
 * Although, it seems more logical to initialize the eecd in 
 * ort_execute_parallel(), this would have to be done serially. What we 
 * do here is defer this initialization to each thread, in order to have 
 * them initialize in parallel.
 */
void ort_ee_dowork(int eeid, void *parent_info)
{
  ort_eecb_t      *t = __MYCB, *parent = (ort_eecb_t *) parent_info;
  ort_task_node_t *task_parent, *new_node;
  
  if (t == NULL)         /* 1st time around */
  {
    __SETMYCB(t = ort_calloc_aligned(sizeof(ort_eecb_t), NULL));
    t->ee_info = NULL;   /* not needed actually due to calloc */

    /* initialize thread's task structures */
    t->task_data = (ort_task_data_t *) ort_calloc(sizeof(ort_task_data_t));
    #if !defined(HAVE_ATOMIC_FAA)
      ee_init_lock((ee_lock_t *) &(t->task_data->lock), ORT_LOCK_NORMAL);
    #endif
  }

  /* Prepare my control block */
  t->parent         = parent;
  t->num_siblings   = parent->num_children;
  t->level          = parent->level + 1;       /* 1 level deeper */
  t->thread_num     = eeid;                   /* Thread id within the team */
  t->num_children   = 0;
  t->shared_data    = NULL;
  t->sdn            = parent;
  t->mynextNWregion = 0;   
  t->nowaitregion   = 0;          /* VVD--actually we don't need to do this */
  t->activelevel    = parent->activelevel +       /* OpenMP 3.0 - team of 1 */
                      ((t->num_siblings > 1) ? 1 : 0);   /* is NOT parallel */

  /* Prepare implicit task */
  t->task_data->tasks_left = 0;

  task_parent             = __CURRTASK(parent);
  assert(task_parent != NULL);
  new_node                = (ort_task_node_t*)
                             ort_task_alloc(t, sizeof(ort_task_node_t));
  new_node->func          = parent->workfunc;
  new_node->task_argument = (void *) t;
  new_node->num_children  = 0;
  new_node->next          = NULL;
  new_node->parent        = task_parent;
  new_node->icvs          = task_parent->icvs;

  #if defined(HAVE_ATOMIC_FAA)
    _faa(&((task_parent)->num_children), 1);
    _faa(&(parent->task_data->tasks_left), 1);
  #else
    ee_set_lock(&((task_parent)->lock));
    (task_parent->num_children)++;
    ee_unset_lock(&((task_parent)->lock));
    ee_set_lock(&((parent->task_data)->lock));
    (parent->task_data->tasks_left)++;
    ee_unset_lock(&((parent->task_data)->lock));
  #endif
  
  __CURRTASK(t) = new_node;

  /* Execute the actual function */
  (*(parent->workfunc))(t);

  /* Cleanup implicit task stuff & help with other tasks */
  #if defined(HAVE_ATOMIC_FAA)
    _faa(&((task_parent)->num_children), -1);
    _faa(&(parent->task_data->tasks_left), -1);
  #else
    ee_set_lock(&((task_parent)->lock));
    (task_parent)->num_children--;
    ee_unset_lock(&((task_parent)->lock));
    ee_set_lock(&((parent->task_data)->lock));
    parent->task_data->tasks_left--;
    ee_unset_lock(&((parent->task_data)->lock));
  #endif

  ort_barrier_me();
  __CURRTASK(t) = task_parent;
  ort_task_free(t, new_node);              /* recycle task node */
}


/* This is called upon entry in a parallel region.
 *   (1) I inquire OTHR for num_threads threads
 *   (2) I set up my eecb fields for my children to use
 *   (3) I create the team
 *   (4) I participate, having acquired a new eecb
 *   (5) I wait for my children to finish and resume my old eecb
 *
 * If num_threads = -1, the team will have icvs.nthreads threads.
 */
void ort_execute_parallel(int num_threads, void *(*func)(void *), void *shared)
{
  ort_eecb_t p, *me = __MYCB;
  int        nthr = 0;

  /*
   * First determine how many threads will be created
   */
   
  if (num_threads <= 0)               /* No num_threads() clause */
    num_threads = __CURRTASK(me)->icvs.nthreads;
    
  if (num_threads > 1)
  {
    if (me->level == 0)               /* 1st level of parallelism */
    {
      nthr = ee_request(num_threads-1, 1);
      if (nthr != num_threads-1 && !__CURRTASK(me)->icvs.dynamic)
      {
        TEAM_FAILURE:
        ort_error(3, "failed to create the requested number (%d) of threads.\n"
                     "   Try enabling dynamic adjustment using either of:\n"
                     "    >> OMP_DYNAMIC environmental variable, or\n"
                     "    >> omp_set_dynamic() call.\n", num_threads);
      }
    }
    else                              /* Nested level */
    {
      if (__CURRTASK(me)->icvs.nested &&  /* is nested parallelism enabled? */
           ((ort->icvs.levellimit == -1) || ort->icvs.levellimit > me->level))
      {
        nthr = ee_request(num_threads-1, me->level + 1);
        /* Now here we have an interpretation problem wrt the OpenMP API,
         * in the case nthr != num_threads-1.
         * Should we breakdown or can we ditch "icvs.nested" and create a
         * 1 thread team?
         * Well, we do a bit of both:
         * - if the ee library returned 0, we do the latter with a warning.
         * - otherwise, we do the former; it is easier since otherwise
         *   we must re-contact the othr library to explain to her that
         *   after all we won't need the threads we requested.
         */
        if (nthr == 0)
          ort_warning("parallelism at level %d disabled due "
                      "to lack of threads.\n   >> Using a team of 1 thread.\n",
                      me->level + 1);   /* GF */
        else
          if (nthr != num_threads-1 && !__CURRTASK(me)->icvs.dynamic)
            goto TEAM_FAILURE;
      }
      else
        nthr = 0;      /* Only me will execute it */
    }
  }

  /*
   * Next, initialize everything needed and create the threads
   */
   
  me->num_children = nthr + 1;
  me->shared_data  = (shared == NULL) ? me->sdn->shared_data : shared;
  me->workfunc     = func;
  ee_barrier_init(&me->barrier, nthr+1);     /* 2009-12-17 */
  init_task_queues(me, nthr);                    /* Prepare for tasks */

  if (nthr != 0)
  {
    if (!me->have_created_team)
    {
      ee_init_lock(&me->copyprivate.lock, ORT_LOCK_SPIN);
      me->have_created_team = 1;
      me->workshare.blocking.inited = 0;
    }
    init_workshare_regions((ort_workshare_t *) &me->workshare);

    FENCE;   /* Just to make sure */

    /* Start the threads (except myself - I am the first child, too).
     * We do not initialize the threads eecb fields because it incurs
     * quite an overhead (especially if "nthr" is not small.
     * Instead we require that ee_create() does it by itself,
     * by calling ort_ee_dowork().
     */
    ee_create(nthr, me->level, me, &me->ee_info);
  }

  /*
   * Participate: I am thread # 0, and I will also call the function
   */

  memset(&p, 0, sizeof(p));        /* Sly bug!!!! */
  p.task_data = ort_task_data_alloc(me);
  __SETMYCB(&p);                   /* Change my key */
  if (ORT_DEBUG) ort_debug_thread("just created team and about to participate");
  ort_ee_dowork(0, me);

  /*
   * All done; destroy the team (me now points to my "parent")
   */

  if (nthr > 0)
    ee_waitall(&me->ee_info);      /* Wait till all children finish */

  me->num_children = 0;
  __SETMYCB(me);                   /* assume my parent eecb */
  ort_task_data_free(me, p.task_data); /* recycle task_data */
}


/* Execute a function only by the running thread. This is only called
 * in the parser-generated code when checking an IF condition at a
 * parallel section. In nested parallelism, the running thread may
 * not be the same as the master thread.
 */
void ort_execute_serial(void *(*func)(void *), void *shared)
{
  ort_eecb_t *me, *sdn;
  void       *shd;
  int        num_sibs;
  
  sdn      = (me = __MYCB)->sdn;   /* Backup */
  num_sibs = me->num_siblings;
  shd      = me->shared_data;
  
  me->shared_data = shared;        /* These are the only fields that are */
  me->sdn = me;                    /* checked & touched by the library */
  me->num_siblings = 1;            /* when the team has 1 member */
  me->level++;
  (*func)((void *) me);
  me->level--;
  
  me->shared_data  = shd;               /* Restore */
  me->num_siblings = num_sibs;
  me->sdn          = sdn;
}


void *ort_get_parent_ee_info(void)
{
  return ( __MYCB->parent->ee_info );
}


void *ort_get_shared_vars(void *me)
{
  return ( ((ort_eecb_t *) me)->sdn->shared_data );
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * THREADPRIVATE STUFF                                               *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/* Returns a pointer to a threadprivate variable; if the variable
 * is referenced for the first time, it creates it.
 *
 * The implementation may not exactly conform to OpenMP v25
 * because we mostly use thread pools and a thread X always retains its
 * threadprivate vars, even if X has id Y != X in a following
 * parallel region (the specs say that the X-th copy of the threadprivate
 * variable will be seen by the X-th thread in following regions).
 * Thus threadprivate variables should better not have values that depend
 * on the value of the id of the thread.
 *
 * But of course, all depends on OTHR. In particular, ee_pthreads1
 * always uses the same id for a thread so no problemo there.
 */
void *ort_get_thrpriv(void **key, int size, void *origvar)
{
  void     *var;
  ee_key_t k;

  if (*key == NULL)    /* Uninitialized key */
  {
    /* Non-initialized key; be careful */
    /* we use the ort->preparation_lock, so as not to define 1 more lock */
    ee_set_lock((ee_lock_t *) &ort->preparation_lock);
      if (*key == NULL)
      {
        ee_key_create((ee_key_t *) key, 0);
        FENCE;
      }
    ee_unset_lock((ee_lock_t *) &ort->preparation_lock);
  }

  k = (ee_key_t) *key;            /* Warning on 64-bits */
  if ((var = ee_getspecific(k)) == NULL)
  { /* 1st reference for this thread */
    if (__MYCB == ort_master)     /* Master thread has the origvar */
      var = origvar;
    else
      if ((var = (void *) malloc(size)) == NULL)
        ort_error(1, "[ort_get_thrpriv]: out of memory\n");
    ee_setspecific(k, var);
  }
  return (var);
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * ATOMIC, CRITICAL, REDUCTION AND COPYPRIVATE SUPPORT               *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
                                     

/* The binding thread set of an atomic region is the set of all threads
 * in the program, not just the threads of the current team (OpenMP V2.5).
 * Thus ort->atomic_lock is global.
 */
void ort_atomic_begin()
{
  ee_set_lock((ee_lock_t *) &ort->atomic_lock);    /* ##### */
}


void ort_atomic_end()
{
  ee_unset_lock((ee_lock_t *) &ort->atomic_lock);  /* ##### */
}


/* As in atomic, the binding set is all the threads in the program
 */
void ort_critical_begin(omp_lock_t *critlock)
{
  /* Because critical locks have external scope, they are initialized
   * to NULL, thus we can safely differentiate between uninitialized
   * and initialized ones.
   */
  if (*critlock == NULL) ort_prepare_omp_lock(critlock, ORT_LOCK_SPIN);
  ee_set_lock((ee_lock_t *) *critlock);
}


void ort_critical_end(omp_lock_t *critlock)
{
  ee_unset_lock((ee_lock_t *) *critlock);
}


void ort_reduction_begin(omp_lock_t *redlock)
{
  /* Because OMPi's parser declares all reduction locks as globals, 
   * they are initialized to NULL, thus we can safely differentiate 
   * between uninitialized and initialized ones.
   */
  if (*redlock == NULL) ort_prepare_omp_lock(redlock, ORT_LOCK_SPIN);
  ee_set_lock((ee_lock_t *) *redlock);
}


void ort_reduction_end(omp_lock_t *redlock)
{
  ee_unset_lock((ee_lock_t *) *redlock);
}


/* The SINGLE onwer initialization for copyprivate data.
 * It creates an array of pointers to its private data.
 */
void ort_broadcast_private(int num, ...)
{
  va_list     ap;
  ort_eecb_t  *me;
  ort_cpriv_t *cp;
  int         i;
  
  if ((me = __MYCB)->num_siblings == 1)  /* Nothing here if I am solo */
    return;
    
  cp = &(me->parent->copyprivate);
  cp->owner   = me->thread_num;
  cp->copiers = me->num_siblings;
  cp->data    = (volatile void **) malloc(num * sizeof(void *));
  
  va_start(ap, num);
  for (i = 0; i < num; i++)
    cp->data[i] = va_arg(ap, void *);
  va_end(ap);
}


/* All threads copy copyprivate date from the SINGLE owner.
 * The arguments are pointer-size pairs.
 */
void ort_copy_private(int num, ...)
{
  va_list     ap;
  int         i;
  void        **from, *arg;
  ort_cpriv_t *cp;
  ort_eecb_t  *me;

  if ((me = __MYCB)->num_siblings == 1)  /* Nothing here if I am solo */
    return;
  
  cp = &(me->parent->copyprivate);
  if ( cp->owner != me->thread_num)  /* I am not the owner */
  {
    va_start(ap, num);
    from = (void **) cp->data;
    for (i = 0; i < num; i++)
    {
      arg = va_arg(ap, void *);
      memcpy(arg, from[i], va_arg(ap, int));
    }
    va_end(ap);
  }
  
#if defined(HAVE_ATOMIC_FAA) && !defined(EE_TYPE_PROCESS)
  i = _faa(&(cp->copiers), -1) - 1;
#else
  ee_set_lock(&cp->lock);
    i = --cp->copiers;
  ee_unset_lock(&cp->lock);
#endif

  if (i == 0)
    free(cp->data);                    /* Free allocated data */
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * A PORTABLE BARRIER USED AS A DEFAULT ONE IN OMPi                  *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/* Is is assumed here that only 1 thread calls this (so as to avoid
 * expensive bookkeeping). Reinitializations cause no harm.
 */
void ort_default_barrier_init(ort_defbar_t *bar, int nthr)
{
  if (nthr > MAX_BAR_THREADS)
    ort_error(1, "barrier cannot support > %d threads; "
                 "change MAX_BAR_THREADS in ort.h\n", MAX_BAR_THREADS);
    
  bar->sense = 0;
  bar->nthr = nthr;
  for (--nthr;  nthr >= 0; nthr--)
    bar->arrived[nthr].value = 0;
}


void ort_default_barrier_destroy(ort_defbar_t *bar)
{
}


void ort_default_barrier_wait(ort_defbar_t *bar, int eeid)
{
  int sense;
  volatile int time = 0; FENCE;

  if (bar->nthr == 1) return;

  sense = !( bar->sense );
  if (eeid > 0)
  {
    bar->arrived[eeid].value = sense;
    FENCE;
    for ( ; (bar->sense != sense); time++)
      if (time == (YIELD_FREQUENTLY))
      {
        time = -1;
        ort_taskwait(1);    /* Help with any of the remaining tasks */
        ee_yield();
      }
  }
  else     /* Let the master do the work */
  {
    for (eeid = 1; eeid < bar->nthr; eeid++)
    {
      for ( ; (bar->arrived[eeid].value != sense); time++)
        if (time == (YIELD_IMMEDIATELY))
        {
          time = -1;
          ort_taskwait(1);
          ee_yield();
        }
    }
    bar->sense = sense;      /* Release them all */
    FENCE;
  }
}


void ort_barrier_me(void)
{
  ort_eecb_t *me = __MYCB;

  if (me->num_siblings == 1)
    return;
  ee_barrier_wait(&me->parent->barrier, me->thread_num);
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * Interoperability;                                                 *
 *    an external calling thread may become an OpenMP master.        *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/* Internal funuction to force an arbitrary eecb; if a new eecb is
 * returned, IT MUST NOT BE SIMPLY free()ed due to its aligned allocation.
 */
void *ort_prepare_my_eecb(void *eecb)
{
  if (eecb == NULL)
    eecb = ort_calloc_aligned(sizeof(ort_eecb_t), NULL);
  initialize_eecb((ort_eecb_t *) eecb);
  __SETMYCB((ort_eecb_t *) eecb);
  return (eecb);
}


/* Any user thread which is not an openmp thread may become an independent
 * OpenMP master. Repeated calls just reset the eecb.
 */
int ompi_makeme_openmp_master()
{
  ort_eecb_t *me = __MYCB;
  
  if (me == NULL)
    ort_prepare_my_eecb(NULL);
  else
  {
    if (me->level) 
      return (-1);       /* Error; called from a non-master openmp thread */
    initialize_eecb(me); /* Reset */
  }
  return (0);
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * UTILITY FUNCTIONS                                                 *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
                                     
                                     
void ort_error(int exitcode, char *format, ...)
{
  va_list ap;

  va_start(ap, format);
  fprintf(stderr, "[ORT error]: ");
  vfprintf(stderr, format, ap);
  va_end(ap);

  exit(exitcode);
}


void ort_warning(char *fmt, ...)
{
  va_list ap;

  va_start(ap, fmt);
  fprintf(stderr, "[ORT warning]: ");
  vfprintf(stderr, fmt, ap);
  va_end(ap);
}


void *ort_alloc(int size)
{
  void *a;

  if ((a = malloc(size)) == NULL)
    ort_error(1, "memory allocation failed\n");
  return (a);
}


void *ort_calloc(int size)
{
  void *a;

  if ((a = calloc(1, size)) == NULL)
    ort_error(1, "memory allocation failed\n");
  return (a);
}


void *ort_alloc_aligned(int size, void **actual)
{
  if (actual == NULL)
  {
    void *tmp;
    
    tmp = ort_alloc(size + CACHE_LINE - 1);
    return ( (void *) ( (((ptrint) (tmp)) + CACHE_LINE - 1) &
                        ((ptrint)(-CACHE_LINE)) ) );
  }
  *actual = ort_alloc(size + CACHE_LINE - 1);
  return ( (void *) ( (((ptrint) (*actual)) + CACHE_LINE - 1) &
                        ((ptrint)(-CACHE_LINE)) ) );
}


void *ort_calloc_aligned(int size, void **actual)
{
  if (actual == NULL)
  {
    void *tmp;
    
    tmp = ort_calloc(size + CACHE_LINE - 1);
    return ( (void *) ( (((ptrint) (tmp)) + CACHE_LINE - 1) &
                        ((ptrint)(-CACHE_LINE)) ) );
  }
  *actual = ort_calloc(size + CACHE_LINE - 1);
  return ( (void *) ( (((ptrint) (*actual)) + CACHE_LINE - 1) &
                        ((ptrint)(-CACHE_LINE)) ) );
}


void *ort_realloc_aligned(int size, void **actual)
{
  if ( (*actual = realloc(*actual, size + CACHE_LINE - 1)) == NULL )
    ort_error(1, "memory reallocation failed\n");
  return ( (void *) ( (((ptrint) (*actual)) + CACHE_LINE - 1) &
                        ((ptrint)(-CACHE_LINE)) ) );
}


#if defined(EE_TYPE_PROCESS)
  void ort_shmalloc(void **p, int size, int upd)
  {
    ee_shmalloc(p, size, upd);

    if (!(*p))
      ort_error(1, "shmalloc failed\n");
  }

  void ort_shmfree(void *p)
  {
    ee_shmfree(p);
  }
#endif


/* This is only called from parser-generate code. */
void ort_fence(void)
{
  FENCE;
}


/* User-program (omp) locks are all defined as void *, including
 * the parser-generated locks for critical and reduction directives.
 * Upon initialization of such a lock, an actual othr lock is
 * allocated and initialized, through the following function.
 */

/* Allocate & initialize a user lock safely
 */
void ort_prepare_omp_lock(omp_lock_t *lock, int type)
{
  void *new;
  
  ee_set_lock((ee_lock_t *) &ort->preparation_lock);
    if (*lock == NULL)
    {
      /* The problem we have here is with non-global user locks.
       * Those are not initialized necessarily to NULL, thus
       * we cannot know if a lock is already initialized or
       * not. We must assume it is not initialized. Otherwise,
       * if many threads try to initialize it, we may end up
       * with dangling malloc()s. Of course, the programmer
       * who lets each thread initialize the same user lock, 
       * is a bad programmer.
       */
#if defined(EE_TYPE_PROCESS)
      ee_init_lock((ee_lock_t *)lock, type);
#else
      new = ort_alloc( sizeof(ee_lock_t) );
      ee_init_lock((ee_lock_t *) new, type);
      FENCE; /* 100% initialized, before been assigned to "lock" */
      *lock = new;
      FENCE;
#endif
    }
  ee_unset_lock((ee_lock_t *) &ort->preparation_lock);
}


/* For internal tests only */
void ort_debug_thread(char *fmt, ...)
{
  va_list ap;
  static ee_lock_t *l;
  ort_eecb_t *t = __MYCB;

#define indent() { int i; for (i = 0; i < t->level; i++) fputs("  ", stderr); }
  if (l == NULL)
    { l = malloc(sizeof(ee_lock_t)); ee_init_lock(l, ORT_LOCK_NORMAL); }
  
  ee_set_lock(l);
    if (t == NULL)
    {
      fprintf(stderr, "  *** I am an uninitialized thread!!\n");
      ee_unset_lock(l);
      return;
    }
    indent(); fprintf(stderr, "( ::%ld::\n", (long int) t);
    indent(); fprintf(stderr, "   |       id = %d\n", t->thread_num);
    indent(); fprintf(stderr, "   |    level = %d\n", t->level);
    indent(); fprintf(stderr, "   | teamsize = %d\n", t->num_siblings);
    indent(); fprintf(stderr, "   |  ee_info = %ld\n", (long int) t->ee_info);
    indent(); fprintf(stderr, "   |   parent = %ld\n", (long int) t->parent);

    va_start(ap, fmt);
    if (fmt)
    {
      fprintf(stderr, "\n");
      indent(); fprintf(stderr, "  MESSAGE:\n");
      indent(); fprintf(stderr, "    "); vfprintf(stderr, fmt, ap);
      indent(); fprintf(stderr, "\n");
    }
    va_end(ap);

    indent(); fprintf(stderr, ")\n");
  ee_unset_lock(l);
#undef indent
}
