/*
  OMPi OpenMP Compiler
  == Copyright since 2001 the OMPi Team
  == Department of Computer Science, University of Ioannina

  This file is part of OMPi.

  OMPi is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  OMPi is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with OMPi; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* ORT_PRIVE.H
 * Definitions and types needed for building ORT.
 * This is only included by ort.c.
 */
 
#ifndef __ORT_PRIVE_H__
#define __ORT_PRIVE_H__

#include "ort.h"
#include <ee.h>

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                       *
 *  CONSTANTS AND MACROS                                 *
 *                                                       *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/* Maximum number of concurrently active nowait workshare regions.
 * A number equal to a small multiple of the # of threads seems good,
 * although the more the better. However, few programs have many
 * nowait regions so we don't choose a huge number (so as to not occupy
 * too much space).
 */
#define MAXACTIVEREGIONS 50

#define _OMP_SINGLE   0        /* Workshare types */
#define _OMP_SECTIONS 1
#define _OMP_FOR      2

#define _OMP_ACTIVE   0        /* Wait policies */
#define _OMP_PASSIVE  1

#if defined(EE_TYPE_PROCESS)

  #define ee_key_t             oprc_key_t
  #define ee_key_create        oprc_key_create
  #define ee_getspecific       oprc_getspecific
  #define ee_setspecific       oprc_setspecific
  
  #define ee_initialize        oprc_initialize
  #define ee_finalize          oprc_finalize
  #define ee_request           oprc_request
  #define ee_create            oprc_create
  #define ee_yield             oprc_yield
  #define ee_waitall           oprc_waitall
  
  #define ee_lock_t            oprc_lock_t
  #define ee_init_lock         oprc_init_lock    
  #define ee_destroy_lock      oprc_destroy_lock 
  #define ee_set_lock          oprc_set_lock     
  #define ee_unset_lock        oprc_unset_lock   
  #define ee_test_lock         oprc_test_lock

  #if defined(AVOID_OMPI_DEFAULT_BARRIER)
    #define ee_barrier_t         oprc_barrier_t
    #define ee_barrier_init      oprc_barrier_init
    #define ee_barrier_destroy   oprc_barrier_destroy
    #define ee_barrier_wait      oprc_barrier_wait
  #endif

  #if defined(AVOID_OMPI_DEFAULT_TASKS)
    #define ee_new_task          oprc_new_task
    #define ee_new_task_exec     oprc_new_task_exec
    #define ee_taskwait          oprc_taskwait
  #endif

#else
  
  #define ee_key_t             othr_key_t
  #define ee_key_create        othr_key_create
  #define ee_getspecific       othr_getspecific
  #define ee_setspecific       othr_setspecific

  #define ee_initialize        othr_initialize
  #define ee_finalize          othr_finalize
  #define ee_request           othr_request
  #define ee_create            othr_create
  #define ee_yield             othr_yield
  #define ee_waitall           othr_waitall

  #define ee_lock_t            othr_lock_t
  #define ee_init_lock         othr_init_lock
  #define ee_destroy_lock      othr_destroy_lock
  #define ee_set_lock          othr_set_lock
  #define ee_unset_lock        othr_unset_lock
  #define ee_test_lock         othr_test_lock
  
  #if defined(AVOID_OMPI_DEFAULT_BARRIER)
    #define ee_barrier_t         othr_barrier_t
    #define ee_barrier_init      othr_barrier_init
    #define ee_barrier_destroy   othr_barrier_destroy
    #define ee_barrier_wait      othr_barrier_wait
  #endif

  #if defined(AVOID_OMPI_DEFAULT_TASKS)
    #define ee_new_task          othr_new_task
    #define ee_new_task_exec     othr_new_task_exec
    #define ee_taskwait          othr_taskwait
  #endif

#endif

#if !defined(AVOID_OMPI_DEFAULT_BARRIER)
  #define ee_barrier_t         ort_defbar_t
  #define ee_barrier_init      ort_default_barrier_init
  #define ee_barrier_destroy   ort_default_barrier_destroy
  #define ee_barrier_wait      ort_default_barrier_wait
#endif


/* This is for the case where some eelib supports nested parallelism
 * but requires dynamic to be turned on.
 */
#define check_nested_dynamic(n,d)\
          if ((n) && !(d) && !ort->eecaps.supports_nested_nondynamic) {\
            ort_warning("the EE library reports that nested and NOT dynamic\n"\
                        "   parallelism cannot be supported.\n"\
                        "   Try enabling dynamic adjustment using either of:\n"\
                        "    >> OMP_DYNAMIC environmental variable, or\n"\
                        "    >> omp_set_dynamic() call.\n\n"\
                   "*** disabling support for nested parallelism for now ***\n"\
                   "[end of ORT warning]\n");\
            (n) = 0;\
          }


/* Busy waiting with yield.
 * We do a FENCE in the beginning for many reasons. One is that
 * usually we wait on stuff other threads may modify. Another
 * is that at least one optimizing compiler optimizes away the
 * loop, incorrectly.
 */
#define YIELD_IMMEDIATELY 0
#define YIELD_FREQUENTLY  50
#define YIELD_OCCASIONALY 150
#define YIELD_RARELY      500
#define OMPI_WAIT_WHILE(f, trials_before_yielding) { \
          volatile int time = 0; FENCE; \
          for ( ; (f); time++) \
            if (time == (trials_before_yielding)) { \
              time = -1; \
              ee_yield(); \
            }; \
        }


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                       *
 *  TYPES                                                *
 *                                                       *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


/* 
 * Types related to tasks
 */


#define POOLSIZE 16
struct _Args;

typedef struct Node
{
    void            *(*func)(void *);  /* Task function */
    struct _Args    *task_argument;    /* Task arguments */
    volatile int    num_children;      /* Number of task's descendants*/
    struct Node     *parent;           /* Task's parent id */
    struct Node     *next;             /* For use in garbage collector */
    #if !defined(HAVE_ATOMIC_CAS)
      ee_lock_t     lock;              /* Lock queue for atomic access */
    #endif
    ort_task_icvs_t icvs;              /* OpenMP3.0 */
} ort_task_node_t;

typedef struct Queue
{
    volatile int top;	    /* Must be volatile in order to read it atomicaly */
    volatile int bottom;
    ort_task_node_t* tasks[POOLSIZE];
    #if !defined(HAVE_ATOMIC_FAA)
      ee_lock_t lock;            /* Lock queue for atomic access */
    #endif
} ort_task_queue_t;

typedef struct _Args
{
	void *function_arguments;     /* Task's function arguments */
	ort_task_node_t* task_parent; /* Parent task structure */
        struct _Args* next;           /* For use in garbage collector */
} ort_task_args_t;

/* Task environment recycle subbin*/
typedef struct task_env_rsubbin ort_task_env_rsubbin_t;
struct task_env_rsubbin {
  ort_task_env_rsubbin_t* next;
} ;

/* Task environment recycle bin */
typedef struct task_env_rbin ort_task_env_rbin_t;
struct task_env_rbin {
  int bin_size;
  ort_task_env_rsubbin_t* sub_bin;
  ort_task_env_rbin_t *next;
};

/* Hold data for task implementation */
typedef struct task_data_s{
  /* Task environment recycle bin */
  ort_task_env_rbin_t* task_env_bin;
  /* Thread's private data */
  /* Have to know what task i currently execute. Needed in task wait */
  ort_task_node_t *current_executing_task;
  /* Recycle bin for task argument */
  ort_task_args_t *task_argument_recycle_bin;
  /* Common data between me and rest threads of the group (my children).*/
  /* Obviusly used when i become father */
  /* How many tasks are left to be done from the members of my group */
  volatile int tasks_left;
  /* table that holds task queues of all threads in my group */
  ort_task_queue_t* queue_table;
  /* Maximum number of children that i have created */
  int max_children;
  /* Recycle task_data_struct. Used only when i become parent due to sly bug */
  struct task_data_s* task_data_recycle_bin;
  ort_task_node_t* task_node_recycle_bin;
  #if !defined(HAVE_ATOMIC_FAA)
    /* Lock queue for atomic access */
    ee_lock_t lock;
  #endif
} ort_task_data_t;


/*
 * Other types
 */


/* Ordered data. */
typedef struct {
  int       next_iteration;  /* Low bound of the chunk that should be next */
  int       incr;            /* Is > 0 if FOR's step is positive */
  ee_lock_t lock;
} ort_ordered_info_t;


/* For FOR loops */
typedef struct {
  /* lb is initialized to the loop's initial lower bound. During execution,
   * it represents the "current" lower bound, i.e. the next iteration to be
   * scheduled.
   * *** IT IS ONLY USED FOR THE GUIDED & DYNAMIC SCHEDULES ***
   */
  volatile int       lb;           /* The next iteration to be scheduled */
  ort_ordered_info_t ordering;      /* Bookeeping for the ORDERED clause */
} ort_forloop_t;


/* For workshare regions */
typedef struct
{
  ee_lock_t    reglock;                         /* Lock for the region */
  volatile int empty;                 /* True if no thread entered yet */
  volatile int notleft;            /* # threads that have not left yet */
  int          inited;              /* 1 if the region was initialized */

  /* SECTIONS & FOR specific data */
  volatile int  sectionsleft; /* Remaining # sections to be given away */
  ort_forloop_t forloop;      /* Stuff for FOR regions */
} wsregion_t;


/* A table of simultaneously active workshare regions */
typedef struct {
  /* This is for BLOCKING (i.e. with no NOWAIT clause) regions. */
  wsregion_t blocking;
  /* This is for keeping track of active NOWAIT regions.  */
  volatile int headregion, tailregion;   /* volatile since all threads watch */
  wsregion_t active[MAXACTIVEREGIONS];
} ort_workshare_t;


/* Holds pointers to copyprivate vars. */
typedef struct {
  volatile void  **data;
  int            owner;
  int            copiers;
  ee_lock_t      lock;
} ort_cpriv_t;


/* Execution entity control block (eecb).
 * ORT keeps such a block for every ee; it contains fields necessary 
 * for runtime bookkeeping.
 * The eecb's form a tree, where child ee's have pointers to their
 * parent's eecb.
 */
typedef struct ort_eecb_s ort_eecb_t;
struct ort_eecb_s {
  /*
   * The barrier is declared first, hoping it will be cache aligned
   */
  
  /* First, the fields used by my children (when I am the parent)
   */ 
  ee_barrier_t    barrier;                         /* Barrier for my children */
  int             have_created_team;  /* 1 if I was a team parent in the past */
  int             num_children;
  void            *(*workfunc)(void *);   /* The func executed by my children */
  ort_workshare_t workshare;     /* Some fields volatile since children snoop */
  ort_cpriv_t     copyprivate;  /* For copyprivate; owner stores data here and
                                   the rest of the children grab it from here */
  
  /* Fields for me, as a member of a team
   */
  ort_eecb_t *parent;
  
  int thread_num;                                /* Thread id within the team */
  int num_siblings;                                   /* # threads in my team */
  int level;                            /* At what level of parallelism I lie */
  int activelevel;             /* At what *active* level of parallelism I lie */
  void *shared_data;          /* Pointer to shared struct of current function */
  ort_eecb_t *sdn;      /* Where I will get shared data from; normally
                                   from my parent, except at a false parallel
                                   where I get it from myself since I am 
                                   the only thread to execute the region. */
  int mynextNWregion;        /* Non-volatile; I'm the only thread to use this */
  int chunklb;            /* Non-volatile; my current chunk's start iteration;
                             changes for every chunk i execute through
                             ort_thischunk_lb(); only used in ordered_begin() */
  int nowaitregion;             /*  True if my current region is a NOWAIT one */

  /* Tasking structures
   */
  ort_task_data_t *task_data;

  /* Thread-library specific data
   */
  void *ee_info;                                 /* Handled by the ee library */

};


/* List holding pointers to user's shared global variables (only if ee=proc) */
typedef struct ort_sglvar_s ort_sglvar_t;
struct ort_sglvar_s {
  void         **varptr;      /* Pointer to user's global variable */  
  int          size;          /* sizeof(var) */
  void         *initvalue;    /* Initializer */
  ort_sglvar_t *next;
};


/* All global variables ORT handles; if ee=proc, this is also placed
 * in shared memory.
 */
typedef struct {
    ort_icvs_t         icvs;
    ort_caps_t         eecaps;
    volatile ee_lock_t atomic_lock;       /* Global lock for atomic */
    volatile ee_lock_t preparation_lock;  /* For initializing user locks */
  } ort_vars_t;


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                       *
 *  VARIABLES & MORE MACROS                              *
 *                                                       *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


extern ort_vars_t *ort;

/* My eecb */
#ifdef USE_TLS
  extern TLS_KEYWORD void *myeecb;
  #define __SETMYCB(v) (myeecb = v)
  #define __MYCB       ((ort_eecb_t *) myeecb)
#else
  extern ee_key_t eecb_key;
  #define __SETMYCB(v) ee_setspecific(eecb_key,v)
  #define __MYCB       ((ort_eecb_t *) ee_getspecific(eecb_key))
#endif

#define __CURRTASK(eecb) ((eecb)->task_data->current_executing_task)

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                         *
 *  FUNCTIONS etc (also used by the parser, see ort.defs)  *
 *                                                         *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


int   ort_initialize(int *argc, char ***argv);
void  ort_finalize(int exitval);
void *ort_calloc(int size);
void  ort_execute_serial(void *(*func)(void *), void *shared);
void  ort_execute_parallel(int numthreads, void *(*func)(void *), void *shared);
void  ort_prepare_omp_lock(omp_lock_t *lock, int type);

void *ort_get_shared_vars(void *);
void *ort_get_thrpriv(void **key, int size, void *origvar);
void  ort_sglvar_allocate(void **dataptr, int size, void *initer);
void  ort_fence();

/* Atomic, critical, reduction, copyprivate and barrier */
void ort_atomic_begin();
void ort_atomic_end();
void ort_critical_begin(void **cl);
void ort_critical_end(void **cl);
void ort_reduction_begin(void **cl);
void ort_reduction_end(void **cl);
void ort_broadcast_private(int num, ...);
void ort_copy_private(int num, ...);
void ort_barrier_me(void);

/*
 * From ort_ws.c
 */

/* Structure to optimize guided/dynamic schedules.
 * We compute it once and reuse it in every call to ort_get_xxx_chunk().
 * Such things are declared by the compiler (as void *), are initialized
 * during ort_entering_for() and are utilized in every call to
 * ort_get_xxx_chunk() (static schedules ignore this, though).
 */
typedef struct _ort_gdopt_
        {
          volatile int  *data;  /* Denotes the current lb of the loop */
          volatile void *lock;  /* Lock to access *data */
          int           nth;    /* # siblings */
          void          *me;    /* my info node */
        } ort_gdopt_t;

  /* Workshare-related functions */
void init_workshare_regions(ort_workshare_t *ws);
int  ort_mysingle(int hasnowait);
void ort_leaving_single();
void ort_entering_sections(int hasnowait, int numberofsections);
void ort_leaving_sections();
int  ort_get_section();
void ort_entering_for(int nowait,int hasordered,int lb,int step,ort_gdopt_t *t);
int  ort_leaving_for();
void ort_ordered_begin();
void ort_ordered_end();
void ort_thischunk_lb(int);

  /* For schedules support */
typedef int (*chunky_t)(int lb, int ub, int step, int chunksize,
                        int *from, int *to, int *extra, ort_gdopt_t *opt);

void ort_get_runtime_schedule_stuff(chunky_t *func, int *chunksize);
int  ort_get_guided_chunk(int lb, int ub, int step, int chunksize,
                         int *from, int *to, int *ignored, ort_gdopt_t *opt);
int  ort_get_dynamic_chunk(int lb, int ub, int step, int chunksize,
                         int *from, int *to, int *ignored, ort_gdopt_t *opt);
int  ort_get_runtimestatic_chunk(int lb, int ub, int step, int chunksize,
                         int *from, int *to, int *chunkid, ort_gdopt_t *opt);
int  ort_get_static_default_chunk(int lb, int ub, int step, int *from, int *to);
void ort_init_static_chunksize(int lb, int ub, int step, int chunksize,
                               int *nchunks, int *init_from);

/*
 * From ort_tasks.c
 */

void             ort_init_tasking();
void             ort_new_task(int untied, void *(*func)(void *arg), void *arg);
void             ort_new_task_exec(int untied, void *(*func)(void *arg), void *arg);
void             ort_taskwait(int waitall);
int              ort_task_throttling();
void            *ort_taskenv_alloc(int size);
void             ort_taskenv_free(void *ptr, int size);
ort_task_data_t *ort_task_data_alloc(ort_eecb_t *me);
void             ort_task_data_free(ort_eecb_t *me, ort_task_data_t* recycled_task_data);
ort_task_node_t *ort_task_alloc(ort_eecb_t *me, int size);
void             ort_task_free(ort_eecb_t *me, ort_task_node_t* node);
void             init_task_queues(ort_eecb_t *me, int nthr);

#endif     /* __ORT_PRIVE_H__ */
