/*
  OMPi OpenMP Compiler
  == Copyright since 2001 the OMPi Team
  == Department of Computer Science, University of Ioannina

  This file is part of OMPi.

  OMPi is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  OMPi is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with OMPi; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* ort_tasks.c -- OMPi RunTime library, tasking */

/*
 * Version 1.0.1j:
 *   first time around, out of ort.c code.
 */
 
#include "ort_prive.h"
#include <stdlib.h>
#include <stdio.h>

#define FAILURE 0
#define SUCCESS 1


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * GLOBAL VARIABLES / DEFINITIONS / MACROS                           *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
                                     

/* Per-thread task throttling */
#ifdef USE_TLS
  TLS_KEYWORD int throttle;  
  #define __start_throttling() (throttle = 1)
  #define __stop_throttling()  (throttle = 0)
  #define __check_throttling() (throttle + 0)
#else
  ee_key_t   throttle_key;    /* For thread-specific task throttling */
  #define __start_throttling() ee_setspecific(throttle_key,(void *) 1)
  #define __stop_throttling()  ee_setspecific(throttle_key,(void *) 0)
  #define __check_throttling() (0 != (int) ee_getspecific(throttle_key))
#endif

#define atomic_read(X) *((int*)X)


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * TASKS                                                             *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
                                     

void ort_init_tasking()
{
#ifdef USE_TLS
#else
  ee_key_create(&throttle_key, 0);
#endif
  __stop_throttling();
}


/* Allocate memory for my group task queues */
void init_task_queues(ort_eecb_t *me, int nthr)
{
  int i;
  
  /* Thread becomes parent for first time */
  if (me->task_data->queue_table == NULL)
  {
    me->task_data->queue_table = (ort_task_queue_t*)
                                 ort_alloc((nthr + 1)*sizeof(ort_task_queue_t));
    for (i=0; i<nthr + 1; i++)   /* Initialize task queues */
    {
      me->task_data->queue_table[i].top         = 0;
      me->task_data->queue_table[i].bottom      = 0;
      #if !defined(HAVE_ATOMIC_FAA)
        ee_init_lock((ee_lock_t *) &(me->task_data->queue_table[i].lock),
                     ORT_LOCK_NORMAL);
      #endif
    }
    me->task_data->max_children = nthr + 1;
  }
  else if (me->task_data->max_children < nthr + 1)  /* realloc needed */
  {
    me->task_data->queue_table = (ort_task_queue_t*)
       realloc(me->task_data->queue_table, (nthr + 1)*sizeof(ort_task_queue_t));
    for (i=0; i<nthr + 1; i++)
    {
      me->task_data->queue_table[i].top    = 0;
      me->task_data->queue_table[i].bottom = 0;
    }
    for (i=me->task_data->max_children; i<nthr + 1; i++)
    {
      #if !defined(HAVE_ATOMIC_FAA)
        ee_init_lock((ee_lock_t *) &(me->task_data->queue_table[i].lock),
                     ORT_LOCK_NORMAL);
      #endif
    }
    me->task_data->max_children = nthr + 1;
  }
}


/* ort_task_data_alloc
 * This function allocates temporary memory for use by groups's master thread
 */
ort_task_data_t *ort_task_data_alloc(ort_eecb_t *me)
{
  ort_task_data_t *recycled_task_data = me->task_data->task_data_recycle_bin;

  if (recycled_task_data != NULL)
    me->task_data->task_data_recycle_bin = NULL;
  else
  {
    recycled_task_data = (ort_task_data_t *)ort_alloc(sizeof(ort_task_data_t));
    recycled_task_data->queue_table               = NULL;
    recycled_task_data->task_argument_recycle_bin = NULL;
    recycled_task_data->max_children              = 0;
    recycled_task_data->task_env_bin              = NULL;
    recycled_task_data->task_data_recycle_bin     = NULL;
    recycled_task_data->task_node_recycle_bin     = NULL;
    #if !defined(HAVE_ATOMIC_FAA)
      ee_init_lock((ee_lock_t *) &(recycled_task_data->lock), ORT_LOCK_NORMAL);
    #endif
  }
  return ( recycled_task_data );
}


void ort_task_data_free(ort_eecb_t *me, ort_task_data_t* recycled_task_data)
{
  me->task_data->task_data_recycle_bin = recycled_task_data;
}


/* ort_task_argument_alloc
 * This function tries to use the recycle bin of process 'thread_id'
 * in order to reuse a previusly used 'ort_task_args_t' structure. if recycle
 * bin is empty then a new 'ort_task_args_t' is returned using malloc.
 */
static 
ort_task_args_t *ort_task_argument_alloc(ort_eecb_t *me)
{
  ort_task_args_t *node = me->task_data->task_argument_recycle_bin;

  if (node == NULL)
    return (ort_task_args_t*) ort_alloc(sizeof(ort_task_args_t));

  me->task_data->task_argument_recycle_bin = node->next;
  return node;
}


/* ort_task_argument_free
 * This function puts a 'ort_task_args_t' structure in process's 'thread_id'
 * recycle bin in order to reuse it in the future.
 */
static 
void ort_task_argument_free(ort_eecb_t *me, ort_task_args_t *targ)
{
  targ->next = me->task_data->task_argument_recycle_bin;
  me->task_data->task_argument_recycle_bin = targ;
}


/* ort_task_alloc
 * This function tries to use the recycle bin of process 'thread_id'
 * in order to reuse a previusly used 'ort_task_node_t' structure. if recycle
 * bin is empty then a new 'ort_task_node_t' is returned using malloc.
 */
ort_task_node_t* ort_task_alloc(ort_eecb_t *me, int size)
{
  ort_task_node_t *gc_bin =
                   me->task_data->task_node_recycle_bin;
  ort_task_node_t *recycled_node = gc_bin;

  if (gc_bin == NULL)
    return (ort_task_node_t*)ort_alloc(size);
  
  gc_bin = gc_bin->next;
  me->task_data->task_node_recycle_bin = gc_bin;
  
  return recycled_node;
}


/* ort_task_free
 * This function puts a 'ort_task_node_t' structure in process's 'thread_id'
 * recycle bin in order to reuse it in the future.
 */
void ort_task_free(ort_eecb_t *me, ort_task_node_t* node)
{
  ort_task_node_t *gc_bin = 
                     me->task_data->task_node_recycle_bin;

  node->next = gc_bin;
  gc_bin = node;
  me->task_data->task_node_recycle_bin = gc_bin;
}


/* ort_task_worker_enqueue
 * This function is used by a process (worker) in order to enqueue a new
 * task_node to its task queue.
 */
static int ort_task_worker_enqueue(int worker_id, void* (*func)(void *),
                                   ort_task_args_t *task_argument)
{

    ort_task_node_t *new_node;
    ort_eecb_t *me = __MYCB;
    ort_eecb_t *my_parent = me->sdn;
    int old_bottom = atomic_read
                       (&(my_parent->task_data->queue_table[worker_id].bottom));
    int old_top = atomic_read
                       (&(my_parent->task_data->queue_table[worker_id].top));
    int size = old_bottom - old_top;

    if (size >= POOLSIZE) /* Task queue full */
    {
        return FAILURE;
    }
    
    new_node = (ort_task_node_t*)ort_task_alloc(me, sizeof(ort_task_node_t));
    
    #if !defined(HAVE_ATOMIC_CAS)
      ee_init_lock((ee_lock_t *) &(new_node->lock), ORT_LOCK_NORMAL); 
    #endif
    new_node->func          = func;
    new_node->task_argument = task_argument;
    new_node->num_children  = 0;
    new_node->next          = NULL;
    new_node->parent        = task_argument->task_parent;
    new_node->icvs          = new_node->parent->icvs;

    if(new_node->parent != NULL)   /* Add a new child to parent task */
    {
      #if defined(HAVE_ATOMIC_FAA)
        _faa(&((new_node->parent)->num_children),1);
      #else
        ee_set_lock(&((new_node->parent)->lock));
        (new_node->parent)->num_children++;
        ee_unset_lock(&((new_node->parent)->lock));
      #endif
    }

    #if defined(HAVE_ATOMIC_FAA)
      _faa(&(my_parent->task_data->tasks_left),1);
    #else
      ee_set_lock(&((my_parent->task_data)->lock));
      my_parent->task_data->tasks_left++;
      ee_unset_lock(&((my_parent->task_data)->lock));
    #endif

    my_parent->task_data->queue_table[worker_id].tasks[old_bottom%POOLSIZE] =
                                                                       new_node;

    #if defined(HAVE_ATOMIC_FAA)
      _faa(&(my_parent->task_data->queue_table[worker_id].bottom),1);
    #else
      ee_set_lock(&(my_parent->task_data->queue_table[worker_id].lock));
      my_parent->task_data->queue_table[worker_id].bottom++;
      ee_unset_lock(&(my_parent->task_data->queue_table[worker_id].lock));
    #endif

    return SUCCESS;
}


/* ort_task_worker_dequeue
 * This function is used by a process (worker) in order to dequeue a
 * new task_node from its task queue.
 */
static ort_task_node_t*  ort_task_worker_dequeue(int worker_id)
{
    ort_task_node_t* extracting_node;
    ort_eecb_t *my_parent = __MYCB->sdn;
    int old_top;
    int new_top;
    int size;

    #if defined(HAVE_ATOMIC_FAA)
      _faa(&(my_parent->task_data->queue_table[worker_id].bottom),-1);
    #else
      ee_set_lock(&(my_parent->task_data->queue_table[worker_id].lock));
      my_parent->task_data->queue_table[worker_id].bottom--;
      ee_unset_lock(&(my_parent->task_data->queue_table[worker_id].lock));
    #endif

    old_top = 
             atomic_read(&(my_parent->task_data->queue_table[worker_id].top));
    new_top = old_top + 1;
    size = atomic_read(&(my_parent->task_data->queue_table[worker_id].bottom))
               - old_top;

    if (size < 0) /* Queue is empty */
    {
        my_parent->task_data->queue_table[worker_id].bottom = old_top;
        return NULL;
    }

    extracting_node = my_parent->task_data->queue_table[worker_id].tasks\
                             [atomic_read(&((my_parent->task_data->queue_table\
                             [worker_id]).bottom))%POOLSIZE];
    if(size > 0)
        return extracting_node;

    /* If there is only one task left in queue... */
    #if defined(HAVE_ATOMIC_CAS)
      /* If a thief stole the last task... */
      if(!_cas(&((my_parent->task_data->queue_table[worker_id]).top)\
                                                , old_top, new_top))
        extracting_node = NULL;/* then return NULL, else return the last task */
    #else
      ee_set_lock(&((my_parent->task_data->queue_table[worker_id]).lock));

      if((my_parent->task_data->queue_table[worker_id]).top == old_top)
        (my_parent->task_data->queue_table[worker_id]).top = new_top;
      else
        extracting_node = NULL;

      ee_unset_lock(&((my_parent->task_data->queue_table[worker_id]).lock));
    #endif

    /* old_top + 1 = new_top */
    my_parent->task_data->queue_table[worker_id].bottom = old_top + 1;

    return extracting_node;
}

/* ort_task_thief_steal
 * This function is used by a process (thief) in order to dequeue a new
 * task_node from a victim process's task queue.
 */

static ort_task_node_t* ort_task_thief_steal(ort_eecb_t *me, int victim_id)
{
    ort_task_node_t *extracting_node;
    ort_eecb_t *my_parent = me->sdn;
    int old_top = atomic_read\
                          (&(my_parent->task_data->queue_table[victim_id].top));
    int new_top = old_top + 1;
    int old_bottom = atomic_read\
                       (&(my_parent->task_data->queue_table[victim_id].bottom));
    int size = old_bottom - old_top;

    if(size <= 0) /* Victim's queue is empty */
        return NULL;
    /* Steal a task from vitim's top! */
    extracting_node = my_parent->task_data->queue_table\
                                            [victim_id].tasks[old_top%POOLSIZE];

    #if defined(HAVE_ATOMIC_CAS)
      /* if thief managed to steal the task... */
      if(_cas(&((my_parent->task_data->queue_table\
                                           [victim_id]).top), old_top, new_top))
        return extracting_node;
    #else
      ee_set_lock(&((my_parent->task_data->queue_table[victim_id]).lock));
      if((my_parent->task_data->queue_table[victim_id]).top == old_top)
      {
        (my_parent->task_data->queue_table[victim_id]).top = new_top;
        ee_unset_lock(&((my_parent->task_data->queue_table[victim_id]).lock));
        return extracting_node;
      }
      ee_unset_lock(&((my_parent->task_data->queue_table[victim_id]).lock));
    #endif

    return NULL;
}


/* ort_task_execute
 * This function is part of openmp interface. When a thread calls it,
 * it dequeues a task from its queue, or steals a task from a
 * victim and executes it.
 */
static int ort_task_execute(ort_eecb_t *me)
{
    int thread_id = me->thread_num;
    ort_eecb_t *my_parent = me->sdn;
    ort_task_node_t* task_to_execute = ort_task_worker_dequeue(thread_id);
    ort_task_node_t* prev_task_to_execute;

    /* Processes task queue is empty! Have to steal something... */
    if(task_to_execute == NULL)
    {
        int victim;

        for(victim=(thread_id+1);victim<(my_parent->num_children+thread_id);victim++)
        {
            task_to_execute = ort_task_thief_steal\
                                             (me, victim%(my_parent->num_children));
             /* If victim's queue is also empty, try next victim */
            if(task_to_execute == NULL)
                    continue;

            /* Finally i stole something and i am about to execute it! */
            /* First update thread executing status... */
            prev_task_to_execute = __CURRTASK(me);
            __CURRTASK(me) = task_to_execute;

            (task_to_execute->func)\
                    (task_to_execute->task_argument->function_arguments);

            /* Finished my job. Update thread executing status */
            __CURRTASK(me) = prev_task_to_execute;

            #if defined(HAVE_ATOMIC_FAA)
              _faa(&(my_parent->task_data->tasks_left), -1);
            #else
              ee_set_lock(&((my_parent->task_data)->lock));
              my_parent->task_data->tasks_left--;
              ee_unset_lock(&((my_parent->task_data)->lock));
            #endif

            if(task_to_execute->parent != NULL)  /* If task has a parent then */
            {
               #if defined(HAVE_ATOMIC_FAA)
                 _faa(&((task_to_execute->parent)->num_children), -1);
               #else
                 ee_set_lock(&((task_to_execute->parent)->lock));
                 (task_to_execute->parent)->num_children--;
                 ee_unset_lock(&((task_to_execute->parent)->lock));
              #endif
            }

            /* Free memory */
            ort_task_argument_free(me, task_to_execute->task_argument);
            ort_task_free(me, task_to_execute); /* Recycle task node */

            return SUCCESS; /* I succesfully stole and executed a task! */
        }
        return FAILURE; /* There was no task left to execute */
    }
    else
    {
        /*Finally i am about to execute a task! */
        /* First update thread executing status... */
        prev_task_to_execute = __CURRTASK(me);
        __CURRTASK(me) = task_to_execute;
        (task_to_execute->func)\
                (task_to_execute->task_argument->function_arguments);

        /* Finished my job. Update thread executing status */
        __CURRTASK(me) = prev_task_to_execute;
        #if defined(HAVE_ATOMIC_FAA)
          _faa(&(my_parent->task_data->tasks_left), -1);
        #else
          ee_set_lock(&((my_parent->task_data)->lock));
          my_parent->task_data->tasks_left--;
          ee_unset_lock(&((my_parent->task_data)->lock));
        #endif

        /* If task has a parent then */
        if(task_to_execute->parent != NULL)
        {
          #if defined(HAVE_ATOMIC_FAA)
            _faa(&((task_to_execute->parent)->num_children), -1);
          #else
            ee_set_lock(&((task_to_execute->parent)->lock));
            (task_to_execute->parent)->num_children--;
            ee_unset_lock(&((task_to_execute->parent)->lock));
          #endif
        }

        /* Free memory */
        ort_task_argument_free(me, task_to_execute->task_argument);
        ort_task_free(me, task_to_execute); /* Recycle task node */
        return SUCCESS; /* I succesfully executed a task! */
    }
}

/* Create a new task
 */
void ort_new_task(int untied, void *(*func)(void *arg), void *arg)
{
#if defined(AVOID_OMPI_DEFAULT_TASKS)

  ee_new_task(untied, func, arg);
  
#else
  ort_eecb_t      *me = __MYCB;
  ort_task_args_t *task_argument;
  ort_task_icvs_t icvs;

  if (me->task_data == NULL)
    ort_error(1,"[ort_new_task]: error in task data initilization!\n");

  if (me->sdn == NULL || me->num_siblings == 1)
  {
    icvs = __CURRTASK(me)->icvs;  /* back them up */
    (*func)(arg);          /* I am the only thread that executes */
    __CURRTASK(me)->icvs = icvs;  /* restore */
    return;
  }

  /* First i have to create the struct that holds new task's arguments */
  task_argument = ort_task_argument_alloc(me);
  task_argument->function_arguments = arg;
  task_argument->task_parent        = __CURRTASK(me);

  if (ort_task_worker_enqueue(me->thread_num, func, task_argument) == FAILURE)
  {
    icvs = __CURRTASK(me)->icvs;  /* back them up */
    __start_throttling();
    (*func)(arg);
    ort_task_argument_free(me, task_argument);
    __CURRTASK(me)->icvs = icvs;  /* restore */
  }
  __stop_throttling();
#endif
}


/* Create a new task & execute it immediately
 */
void ort_new_task_exec(int untied, void *(*func)(void *arg), void *arg)
{
#if defined(AVOID_OMPI_DEFAULT_TASKS)
  ee_new_task_exec(untied, func, arg);
#else
  ort_eecb_t      *me = __MYCB;
  ort_task_icvs_t icvs = __CURRTASK(me)->icvs;  /* back them up */
  (*func)(arg);
  __CURRTASK(me)->icvs = icvs;  /* restore */
#endif
}


/* Taskwait
 */
void ort_taskwait(int waitall)
{
#if defined(AVOID_OMPI_DEFAULT_TASKS)
  ee_taskwait(1);
#else
  ort_eecb_t *me = __MYCB;

  if (me->sdn == NULL || me->num_siblings == 1) return;

  if (waitall)   /* help with all the tasks in current team */
    while (ort_task_execute(me))
      ;
  else           /* execute till all my child tasks finish */
    while (__CURRTASK(me)->num_children > 0)
      ort_task_execute(me);
#endif
}


/* Task throttling.
 * For the moment, this is a per-thread flag that should be adjusted
 * adaptively. 
 * A simple policy would be to __start_throttling() when the number of 
 * tasks in my private queue exceeds c*N where c is a constant and N 
 * is the number of processors. If later I discover that the number
 * fell below this threshold, I __stop_throttling().
 * 
 */
int ort_task_throttling()
{
  return __check_throttling();
}


void *ort_taskenv_alloc(int size)
{
#if defined(AVOID_OMPI_DEFAULT_TASKS)
  return ( malloc(size) );
#else
  ort_task_env_rbin_t *task_env_rec_bin = __MYCB->task_data->task_env_bin;

  if (size<sizeof(ort_task_env_rsubbin_t*))
	size=sizeof(ort_task_env_rsubbin_t*);

  if(task_env_rec_bin == NULL)/* Recycle bin of task_env is empty. Use malloc */
    return ( ort_alloc(size) );

  /* There is a rbin of unknown size, search for desirable rbin size */

  while(task_env_rec_bin != NULL)
  {
    /* I found a subbin with matching size! */
    if(task_env_rec_bin->bin_size == size)
    {
      if(task_env_rec_bin->sub_bin != NULL)
      {
        void* recyecled_task_env = (void*)(task_env_rec_bin->sub_bin);
        task_env_rec_bin->sub_bin = (ort_task_env_rsubbin_t*)\
                                              (task_env_rec_bin->sub_bin->next);
        return recyecled_task_env;
      } /* Subbin is empty return malloc */
      return ( ort_alloc(size) );
    }
    task_env_rec_bin = task_env_rec_bin->next;
  }

  /* I did not find a matching rbin. Use malloc */
  return ( ort_alloc(size) );
#endif
}


void  ort_taskenv_free(void *ptr, int size)
{
#if defined(AVOID_OMPI_DEFAULT_TASKS)
  free(ptr);

  return;
#else
  ort_eecb_t          *me = __MYCB;
  ort_task_env_rbin_t *rbin = me->task_data->task_env_bin;
  ort_task_env_rbin_t *new_env_rbin;
  
  while(rbin != NULL)  /* Search for matching recycle bin */
  {
    if(rbin->bin_size == size) /* I found a rbin with matching size! */
    {
      ((ort_task_env_rsubbin_t*)ptr)->next = rbin->sub_bin;
      rbin->sub_bin = (ort_task_env_rsubbin_t*)ptr;
      return;
    }
    rbin = rbin->next;
  }

  /* I did not find a matching rbin or Recycle bin of task_env is empty.*/
  /* Now i must create one new rbin and a new rsubbin */

  new_env_rbin = (ort_task_env_rbin_t*)ort_alloc(sizeof(ort_task_env_rbin_t));

  new_env_rbin->bin_size = size;
  new_env_rbin->next = me->task_data->task_env_bin;
  me->task_data->task_env_bin = new_env_rbin;

  new_env_rbin->sub_bin = (ort_task_env_rsubbin_t*)ptr;
  new_env_rbin->sub_bin->next = NULL;

  return;
#endif
}
