/*
  OMPi OpenMP Compiler
  == Copyright since 2001 the OMPi Team
  == Department of Computer Science, University of Ioannina

  This file is part of OMPi.

  OMPi is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  OMPi is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with OMPi; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* ort.c -- OMPi RunTime library, workshare regions support */

/*
 * Version 1.0.1j:
 *   first time around, out of ort.c code.
 */
 
#include "ort_prive.h"
#include <stdlib.h>


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * WORKSHARE REGIONS SUPPORT (SINGLE/SECTIONS/FOR)                   *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
                                     

/*
 * The inner workings of what follows is documented seperately in
 * the OMPi docs/ directory.
 */

/* The array of counters is used cyclically */
#define REGION(idx)  active[(idx) % MAXACTIVEREGIONS]


/* Called when a new team is created, so as to initialize
 * the workshare regions support (in the parent of the team).
 * For speed, we only initialize the 1st nowait region;
 * the next region is initialized by the first thread to enter the
 * previous one.
 * It is completely useless if the team has 1 thread.
 */
void init_workshare_regions(ort_workshare_t *ws)
{
  if (!ws->blocking.inited)
  {
    ee_init_lock(&ws->blocking.reglock, ORT_LOCK_SPIN);
    ee_init_lock(&ws->blocking.forloop.ordering.lock, ORT_LOCK_SPIN);
    ws->blocking.inited = 1;
  }
  ws->blocking.empty = 1;
    
  if (!ws->REGION(0).inited)
  {
    ee_init_lock(&ws->REGION(0).reglock, ORT_LOCK_SPIN);
    ee_init_lock(&ws->REGION(0).forloop.ordering.lock, ORT_LOCK_SPIN);
    ws->REGION(0).inited = 1;
  }
  ws->REGION(0).empty = 1;
  ws->headregion = ws->tailregion = 0;
}


/* Enter a workshare region, update the corresponding counters
 * and return 1 if i am the first thread to enter the region.
 * It is guaranteed (by the parser and the runtime routines)
 * that we are within a parallel region with > 1 threads.
 */
static
int enter_workshare_region(ort_eecb_t *me,
                           int wstype, int nowait, int hasordered,
                           int arg1,  /* = # sections if wstype is SECTIONS
                                         = for loop's lower bound if FOR */
                           int arg2)  /* = the loop's step increment if FOR */
{  
  ort_workshare_t *ws;
  wsregion_t      *r;
  int             myreg, notfirst = 1;
  
  ws = &(me->parent->workshare);
  myreg = me->mynextNWregion;
  r = (nowait) ? &( ws->REGION(myreg) ) : &( ws->blocking );
  
  me->nowaitregion = nowait;  /* Remember the status of this region */

  if (nowait)
  {
    /* If too many workshare regions are active, we must block here 
     * until the tail has advanced a bit. 
     */
    OMPI_WAIT_WHILE(ws->headregion - ws->tailregion == MAXACTIVEREGIONS  &&
                    myreg == ws->headregion, YIELD_FREQUENTLY);
    (me->mynextNWregion)++;    /* Make me ready for the next region */
  }
      
  if (!r->empty) return (0);             /* We are not the first for sure */
  
  ee_set_lock(&r->reglock);       /* We'll be short */
    if (r->empty)                           /* I am the first to enter */
    {
      if (wstype == _OMP_SECTIONS)          /* Initialize */
        r->sectionsleft = arg1;
      else
        if (wstype == _OMP_FOR)
        /*  first_to_arrive_at_for(&r->forloop, hasordered,arg1,arg2); */
        {
          r->forloop.lb = arg1;
          if (hasordered)
          {
            r->forloop.ordering.next_iteration = arg1;
            r->forloop.ordering.incr = arg2;
          }
        }
      
      r->notleft = me->num_siblings;      /* None left yet */

      /* Now, prepare/initialize the next region, if not already inited.
       * This is done so as to avoid serially initializing all regions when
       * the team was created
       */
      if (nowait && myreg+1 < MAXACTIVEREGIONS)
      {
        if (!ws->REGION(myreg+1).inited)
        {
          ee_init_lock(&ws->REGION(myreg+1).reglock, ORT_LOCK_SPIN);
          ee_init_lock(&ws->REGION(myreg+1).forloop.ordering.lock,ORT_LOCK_SPIN);
          ws->REGION(myreg+1).inited = 1;
        }
        ws->REGION(myreg+1).empty = 1;
      }
      
      /* Do this last to avoid races with threads that test without locking */
      if (nowait) (ws->headregion)++;                   /* Advance the head */
      FENCE;
      r->empty = notfirst = 0;
    } 
  ee_unset_lock(&r->reglock);

  return ( !notfirst );
}
  

/* Returns the # of threads that have not yet left the region.
 * A zero means that I am the last thread to leave.
 */
static 
int leave_workshare_region(ort_eecb_t *me)
{
  ort_workshare_t *ws;
  wsregion_t      *r;
  int             myreg, remain;

  ws = &(me->parent->workshare);
  myreg = me->mynextNWregion - 1;   /* It is ahead by 1 */
  r = (me->nowaitregion) ? &( ws->REGION(myreg) ) : &( ws->blocking );
  
#if defined(HAVE_ATOMIC_FAA) && !defined(EE_TYPE_PROCESS)
  remain = _faa(&(r->notleft), -1) - 1;
  if (!remain)
  {
    r->empty = 1;
    if (me->nowaitregion)  /* no race here; no need for the following:    */
      (ws->tailregion)++;  /* _faa(&(ws->tailregion), 1); */
  }
#else
  /* Well, it would be better if we used a seperate lock for leaving */
  ee_set_lock(&r->reglock);
    remain = --(r->notleft);            /* Increment the # threads that left */
    if (!remain)                        /* If I am the last to leave */
    {                                
      r->empty = 1;                     /* The region is now empty */
      if (me->nowaitregion)
        (ws->tailregion)++;             /* Advance the tail */
    }
  ee_unset_lock(&r->reglock);
#endif
  
  return (remain);
}

  
/* Returns 1 if the current thread should execute the SINGLE block
 */
int ort_mysingle(int nowait)
{
  ort_eecb_t *me = __MYCB;
  if (me->num_siblings == 1)
    return (1);
  else
    return ( enter_workshare_region(me, _OMP_SINGLE, nowait, 0, 0, 0) );
}


void ort_leaving_single()
{
  ort_eecb_t *me = __MYCB;
  if (me->num_siblings != 1)
    leave_workshare_region(me);
}


/* This is called in the parser-generated code *only* if we are in a
 * parallel region AND the number of threads is > 1.
 */
void ort_entering_sections(int nowait, int numberofsections)
{
  enter_workshare_region(__MYCB,_OMP_SECTIONS,nowait,0,numberofsections,0);
}


/* Ditto */
void ort_leaving_sections()
{
  leave_workshare_region(__MYCB);
}


/* Returns the id of the next section to execute.
 * Returns < 0 if no more sections left.
 * This is guaranteed (by the parser) to be called only if we are
 * within a PARALLEL region with > 1 threads.
 */
int ort_get_section()
{
  wsregion_t *r;
  int        s;
  ort_eecb_t *me = __MYCB;

  s = me->mynextNWregion - 1;  /* My region ('cause it is 1 ahead) */
  r = (me->nowaitregion) ?
           &(me->parent->workshare.REGION(s) ) :
           &(me->parent->workshare.blocking );
  
  if (r->sectionsleft < 0) return (-1);

#if defined(HAVE_ATOMIC_FAA) && !defined(EE_TYPE_PROCESS)
    s = _faa(&(r->sectionsleft), -1) - 1;
#else
  ee_set_lock(&r->reglock);
    s = --(r->sectionsleft);
  ee_unset_lock(&r->reglock);
#endif

  return (s);
}


/* This is called by the parser-generated code, even if we are
 * not in a parallel region.
 */
void ort_entering_for(int nowait, int hasordered, int lb, int incr,
                      ort_gdopt_t *t)
{
  ort_eecb_t *me = __MYCB;

  /* Check if we are not in a parallel region or the team has only 1 thread */
  if (me->num_siblings == 1)
  {
    t->me  = (void *) me;
    t->nth = 1;
    return;
  }
  
  enter_workshare_region(me, _OMP_FOR, nowait, hasordered, lb, incr);

  if (me->nowaitregion)
  {
    /* 1 less since mynextNWregion has been increased upon entrance */
    t->data = &(me->parent->workshare.
                 REGION(me->mynextNWregion - 1).forloop.lb);
    t->lock = (void *) &(me->parent->workshare.
                 REGION(me->mynextNWregion - 1).reglock);
  }
  else
  {
    t->data = &(me->parent->workshare.blocking.forloop.lb);
    t->lock = (void *) &(me->parent->workshare.blocking.reglock);
  }
  t->nth = me->num_siblings;
  t->me = me;
}


int ort_leaving_for()
{
  ort_eecb_t *me = __MYCB;
  if ( me->num_siblings == 1)
    return (0);
  else
    return ( leave_workshare_region(me) );
}


/*
 * ORDERED
 */

void ort_thischunk_lb(int lb)
{
  __MYCB->chunklb = lb;
}


void ort_ordered_begin()
{
  ort_ordered_info_t *o;
  int                mychunklb;
  ort_eecb_t      *me = __MYCB;

  if (me->num_siblings == 1)
    return;
    
  mychunklb = me->chunklb;   /* Get my chunk's lb */
  o = (me->nowaitregion) ?
      /* 1 less since mynextNWregion has been increased upon entrance */
      &(me->parent->workshare.REGION(me->mynextNWregion - 1).forloop.ordering) :
      &(me->parent->workshare.blocking.forloop.ordering);

  /* Busy wait (we could make it more sophisticated) */
  if (o->incr > 0)
    { OMPI_WAIT_WHILE(mychunklb > o->next_iteration, YIELD_FREQUENTLY); }
  else
    { OMPI_WAIT_WHILE(mychunklb < o->next_iteration, YIELD_FREQUENTLY); }
}


void ort_ordered_end()
{
  ort_ordered_info_t *o;
  ort_eecb_t      *me = __MYCB;

  if (me->num_siblings == 1)
    return;
    
  o = (me->nowaitregion) ?
      /* 1 less since mynextNWregion has been increased upon entrance */
      &(me->parent->workshare.REGION(me->mynextNWregion - 1).forloop.ordering) :
      &(me->parent->workshare.blocking.forloop.ordering);
  
#if defined(HAVE_ATOMIC_FAA) && !defined(EE_TYPE_PROCESS)
  _faa(&(o->next_iteration), o->incr);
#else
  ee_set_lock(&o->lock);
    o->next_iteration += o->incr;
  ee_unset_lock(&o->lock);  /* Well, here we may need a fence (?) */
#endif
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                                                                   *
 * FOR SCHEDULES (dynamic, guided, static and runtime)               *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
                                     

/*
 * The inner workings of what follows is documented seperately in
 * the OMPi docs/ directory.
 */


int ort_get_dynamic_chunk(int lb, int ub, int step, int chunksize, 
                          int *from, int *to, int *ignored, ort_gdopt_t *t)
{
  int newlb;
  
  if (chunksize <= 0) ort_error(1,"fatal: dynamic chunksize <= 0 requested!\n");
  
  if (((ort_eecb_t *) t->me)->num_siblings == 1)
  {
    /* Here we are based on the fact that the parser-generated code, checks
     * for < or > to, not <= or >=.
     */
    if (*from == lb && *to == ub)    /* We did our sole chunk */
      return (0);
    *from = lb;                    /* Get just 1 chunk: all iterations */
    *to   = ub;
    return (1);
  }
  
#if defined(HAVE_ATOMIC_FAA) && !defined(EE_TYPE_PROCESS)
  newlb = _faa(t->data, step*chunksize);
#else
  ee_set_lock((ee_lock_t *) t->lock);
    newlb = *(t->data);
    (*(t->data)) += step*chunksize;
  ee_unset_lock((ee_lock_t *) t->lock);
#endif

  if (step >= 0) {
    if (newlb < ub) {
      *from = newlb;
      *to = newlb + chunksize*step;
      if (*to > ub)
        *to = ub;
      return 1;
    }
    else return 0;
  }
  else {
    if (newlb > ub) {
      *from = newlb;
      *to = newlb + chunksize*step;
      if (*to < ub)
        *to = ub;
      return 1;
    }
    else return 0;
  }
}


/* SUN suggests dividing the number of remaining iters by 2; they
 * have found it to work well in general!
 */
int ort_get_guided_chunk(int lb, int ub, int step, int chunksize, 
                         int *from, int *to, int *ignored, ort_gdopt_t *t)
{
  int   newlb, iters;
  div_t d;
  
  if (chunksize <= 0) ort_error(1, "fatal: guided chunksize <= 0 requested!\n");
  
  if (((ort_eecb_t *) t->me)->num_siblings == 1)
  {
    /* Here we are based on the fact that the parser-generated code, checks
     * for < or > to, not <= or >=.
     */
    if (*from == lb && *to == ub)    /* We did our sole chunk */
      return (0);
    *from = lb;                    /* Get just 1 chunk: all iterations */
    *to   = ub;
    return (1);
  }
  
#if defined(HAVE_ATOMIC_CAS) && !defined(EE_TYPE_PROCESS)
  do {
    newlb = *t->data;
    d = div(ub - newlb, step * t->nth);
    iters = d.quot;
    if (d.rem != 0)
      iters++;
    if (iters > chunksize)
      chunksize = iters;
  } while( !_cas(t->data, newlb, newlb + step * chunksize) );
#else
  ee_set_lock((ee_lock_t *) t->lock);
  newlb = *t->data;
  d = div(ub - newlb, step * t->nth);  /* Remaining # iterations / # threads */
  iters = d.quot;
  if (d.rem != 0)
    iters++;
  if (iters > chunksize)               /* Give no less than chunksize iters */
    chunksize = iters;

  (*t->data) += step * chunksize;
  ee_unset_lock((ee_lock_t *) t->lock);
#endif

  if (step >= 0) {
    if (newlb < ub) {
      *from = newlb;
      *to = *from + chunksize*step;
      if (*to > ub)
        *to = ub;
      return 1;
    }
    else return 0;
  }
  else {
    if (newlb > ub) {
      *from = newlb;
      *to = *from + chunksize*step;
      if (*to < ub)
        *to = ub;
      return 1;
    }
    else return 0;
  }
}


/* Return the sole chunk a thread gets assigned
 */
int ort_get_static_default_chunk(int lb, int ub, int step, int *from,int *to)
{
  ort_eecb_t *me = __MYCB;
  div_t      diters, dchunksize;
  int        chunksize, iters, nchunks;
  int        N = me->num_siblings;
  int        myid = me->thread_num;

  if (N == 1)
  {
    *from = lb;
    *to   = ub;
    return ( *from != *to );
  }
    
  diters = div(ub - lb, step);               /* total number of iterations */
  iters = diters.quot;
  if (diters.rem != 0) iters++;

  if (iters <= N)
  {
    nchunks = iters; 
    chunksize = 1;
  }
  else
  {
    nchunks = N;
    dchunksize = div(iters, N);
    chunksize = dchunksize.quot;                 /* iterations in a chunk */
    iters = dchunksize.rem;
    if (iters) chunksize++;     /* first iters threads get this chunksize */
  }

  if (myid < nchunks)
  {
    if (myid < iters || iters == 0)       /* I get a full chunk */
    {                                 
      *from = lb + chunksize*step*myid;
      *to = *from + chunksize*step;
    }
    else                                  /* I get a smaller chunk */
    {      
      *from = lb + chunksize*step*iters + (chunksize - 1)*step*(myid - iters);
      *to = *from + (chunksize - 1)*step;
    }
  }
  else   /* otherwise, there is nothing for me */
    *from = *to = 0;                      /* force no computations */
  
  return ( *from != *to ); 
}


/* Do some initial calculations for the STATIC schedule with chunksize
 * specified.
 */
void ort_init_static_chunksize(int lb, int ub, int step, int chunksize, 
                               int *nchunks, int *init_from)
{
  div_t diters, dnchunks;
  int iters;

  if (chunksize <= 0) ort_error(1, "fatal: static chunksize <= 0 requested!\n");
  diters = div(ub - lb, step);
  iters = diters.quot;
  if (diters.rem != 0)
    iters++;

  dnchunks = div(iters, chunksize);
  *nchunks = dnchunks.quot;
  if (dnchunks.rem != 0)
    (*nchunks)++;

  *init_from = lb + __MYCB->thread_num*step*chunksize;
}


/* Runtime version of the static schedule (suboptimal but unavoidable)
 * chunkid MUST be initialy equal to 0!!
 */
int ort_get_runtimestatic_chunk(int lb, int ub, int step, int chunksize, 
                               int *from, int *to, int *chunkid, ort_gdopt_t *t)
{
  if (((ort_eecb_t *) t->me)->num_siblings == 1)
  {
    if (*chunkid) return (0);      /* Only 1 chunk will be given! */
    (*chunkid)++;
    *from = lb;                    /* Get just 1 chunk: all iterations */
    *to   = ub;
    return (1);
  }

  if (chunksize == -1) /* No chunksize given */
  {
    if (*chunkid) return (0);     /* Only 1 chunk will be given! */
    (*chunkid)++;
    return ( ort_get_static_default_chunk(lb, ub, step, from, to) );
  }
  else                 /* chunksize given */
  {
    int nchunks, initfrom;
    
    ort_init_static_chunksize(lb, ub, step, chunksize, &nchunks, &initfrom);
    if (*chunkid > nchunks) return (0);
    *from = initfrom + chunksize*step*(*chunkid);
    *to = *from + chunksize*step;
    if (((step >= 0) && (*to > ub)) || ((step < 0) && (*to < ub)))
      *to = ub;
    (*chunkid) += t->nth;
    return ( *from != *to );
  }
}


/* This returns the required function & chunksize to support the
 * RUNTIME schedule code.
 */
void ort_get_runtime_schedule_stuff(chunky_t *func, int *chunksize)
{
  ort_eecb_t *me = __MYCB;
  *chunksize = __CURRTASK(me)->icvs.rtchunk;  /* -1 if not given */
  switch ( __CURRTASK(me)->icvs.rtschedule )
  {
    case omp_sched_dynamic:
      *func = ort_get_dynamic_chunk;
      if (*chunksize == -1) *chunksize = 1;
      break;
    case omp_sched_guided:
      *func = ort_get_guided_chunk;
      if (*chunksize == -1) *chunksize = 1;
      break;
    case omp_sched_auto:
      /* The auto schedule is the static one for now */
      *func = ort_get_runtimestatic_chunk;
      if (*chunksize == -1) *chunksize = 1;
      break;
    default:
      *func = ort_get_runtimestatic_chunk;
      break;
  }
}
